import base64
import hashlib

import minimalkv
import pytest

from plateau.utils.store import (
    _azure_bbs_content_md5,
    _azure_cc_content_md5,
    _has_azure_bbs,
    _has_azure_cc,
    copy_keys,
)

STORE_TYPES = ["azure", "fs", "memory"]

try:
    from azure.storage.blob import __version__ as asb_version

    asb_major_version = int(asb_version.split(".")[0])
except ImportError:
    asb_major_version = 0


def _gen_store(store_type, tmpdir, suffix, azure_store_cfg_factory):
    if store_type == "azure":
        cfg = azure_store_cfg_factory(suffix)
    elif store_type == "fs":
        cfg = {"type": "hfs", "path": tmpdir.join(suffix).strpath}
    elif store_type == "memory":
        cfg = {"type": "hmemory"}
    else:
        raise ValueError(f"Unknown store type: {store_type}")

    store = minimalkv.get_store(**cfg)
    for k in store.keys():
        store.delete(k)

    yield store

    for k in store.keys():
        store.delete(k)

    # prevent ResourceWarning
    if hasattr(store, "block_blob_service"):
        store.block_blob_service.request_session.close()
    if hasattr(store, "blob_container_client"):
        store.blob_container_client.close()


@pytest.fixture(params=STORE_TYPES)
def store(request, tmpdir, azure_store_cfg_factory):
    yield from _gen_store(request.param, tmpdir, "store", azure_store_cfg_factory)


@pytest.fixture(params=STORE_TYPES)
def store2(request, tmpdir, azure_store_cfg_factory):
    yield from _gen_store(request.param, tmpdir, "store2", azure_store_cfg_factory)


@pytest.mark.skipif(
    asb_major_version != 2,
    reason=f"Test is specific for azure-storage-blob ~= 2, but detected version {asb_major_version}",
)
def test_azure_implementation(azure_store_cfg_factory):
    cfg = azure_store_cfg_factory("ts")
    store = minimalkv.get_store(**cfg)
    assert _has_azure_bbs(store)
    content = b"foo"
    store.put("key0", content)
    assert (
        base64.b64decode(
            _azure_bbs_content_md5(store.block_blob_service, store.container, "key0")
        ).hex()
        == hashlib.md5(content).hexdigest()
    )


@pytest.mark.skipif(
    asb_major_version < 12,
    reason=f"Test is specific for azure-storage-blob >= 12, but detected version {asb_major_version}",
)
def test_azure12_implementation(azure_store_cfg_factory):
    cfg = azure_store_cfg_factory("ts")
    store = minimalkv.get_store(**cfg)
    assert _has_azure_cc(store)
    content = b"foobar"
    store.put("key0", content)
    assert (
        _azure_cc_content_md5(store.blob_container_client, "key0").hex()
        == hashlib.md5(content).hexdigest()
    )
    assert (
        _azure_cc_content_md5(store.blob_container_client, "key_does_not_exist", True)
        is None
    )


class TestCopy:
    def test_all(self, store, store2):
        store.put("ka", b"ka")
        store.put("kb", b"kb")

        copy_keys({"ka", "kb"}, store, store2)

        assert set(store2.keys()) == {"ka", "kb"}
        assert store2.get("ka") == b"ka"
        assert store2.get("kb") == b"kb"

    def test_part(self, store, store2):
        store.put("ka", b"ka")
        store.put("kb", b"kb")
        store.put("kc", b"kc")

        copy_keys({"ka", "kb"}, store, store2)

        assert set(store2.keys()) == {"ka", "kb"}
        assert store2.get("ka") == b"ka"
        assert store2.get("kb") == b"kb"

    def test_overwrite(self, store, store2):
        store.put("ka", b"ka")

        store2.put("ka", b"la")
        store2.put("kb", b"lb")

        copy_keys({"ka"}, store, store2)

        assert store2.get("ka") == b"ka"
        assert store2.get("kb") == b"lb"

    def test_same(self, store):
        store.put("ka", b"ka")
        store.put("kb", b"kb")

        copy_keys({"ka", "kb"}, store, store)

        assert set(store.keys()) == {"ka", "kb"}
        assert store.get("ka") == b"ka"
        assert store.get("kb") == b"kb"

    def test_missing(self, store, store2):
        store.put("ka", b"ka")

        with pytest.raises(KeyError) as exc:
            copy_keys({"ka", "kb"}, store, store2)
        assert str(exc.value) == "'kb'"

    def test_weird1(self, store, store2):
        k = "k / =+"
        store.put(k, b"foo")

        copy_keys({k}, store, store2)

        assert set(store2.keys()) == {k}
        assert store2.get(k) == b"foo"

    def test_weird2(self, store, store2):
        # regression found during early customer tests
        k = "cube_main++completeness/table/PARTITION_DATE=2018-06-29%2000%3A00%3A00/PARTITION_ID=366/KLEE_TS=2019-02-27%2012%3A59%3A37.904401/9be13a70302441b2aee5bed2af781be7.parquet"
        store.put(k, b"foo")

        copy_keys({k}, store, store2)

        assert set(store2.keys()) == {k}
        assert store2.get(k) == b"foo"

    def test_illegal(self, store, store2):
        with pytest.raises(ValueError) as exc:
            copy_keys({"kö"}, store, store2)
        assert str(exc.value) == "Illegal key: kö"



# noqa: E402
# isort:skip_file

import datetime
import os
import sys
from collections import OrderedDict
from functools import partial

import pandas as pd
import pytest
from minimalkv import get_store_from_url

# fmt: off
pytest.register_assert_rewrite("plateau.io.testing")
# fmt: on

from plateau.core.common_metadata import make_meta  # noqa: E402
from plateau.core.factory import DatasetFactory  # noqa: E402
from plateau.core.testing import (  # noqa: E402
    TIME_TO_FREEZE,
    cm_frozen_time,
    get_dataframe_alltypes,
    get_dataframe_not_nested,
)
from plateau.io_components.metapartition import (  # noqa: E402
    MetaPartition,
    gen_uuid,
    parse_input_to_metapartition,
)
from plateau.core.utils import lazy_store  # noqa: E402
from plateau.io_components.write import store_dataset_from_partitions  # noqa: E402
from plateau.serialization import ParquetSerializer  # noqa: E402


pd.options.mode.chained_assignment = "raise"


@pytest.fixture
def frozen_time():
    """Depend on this fixture to set the time to TIME_TO_FREEZE by patching
    plateau.core._time.* with mock objects.

    Note: you only need one of the fixtures `frozen_time`,
    `distributed_frozen_time`, or `frozen_time_em`:

      * if your test function takes a `execution_mode` parameter, use `frozen_time_em`. It will behave
        like `distributed_frozen_time` if `execution_mode` starts with "dask", and like `frozen_time` otherwise.
      * otherwise, if you are testing for dask/distributed, use `distributed_frozen_time`. Note
        that this includes the effects of `frozen_time`.
      * otherwise, use `frozen_time`
    """
    with cm_frozen_time(TIME_TO_FREEZE):
        yield


@pytest.fixture
def frozen_time_em(frozen_time):
    yield


@pytest.fixture
def df_serializer():
    return ParquetSerializer()


@pytest.fixture(scope="session")
def store_session_factory(tmpdir_factory):
    path = tmpdir_factory.mktemp("fsstore_test")
    path = path.realpath()
    url = f"hfs://{path}"
    return lazy_store(url)


@pytest.fixture
def store_factory(tmpdir):
    path = tmpdir.join("store").strpath
    url = f"hfs://{path}"
    return lazy_store(url)


@pytest.fixture
def store_factory2(tmpdir):
    path = tmpdir.join("store2").strpath
    url = f"hfs://{path}"
    return lazy_store(url)


@pytest.fixture(scope="session")
def store_session(store_session_factory):
    return store_session_factory()


@pytest.fixture
def store(store_factory):
    return store_factory()


@pytest.fixture
def store2(store_factory2):
    return store_factory2()


def _check_and_delete(key, store, delete_orig):
    if key not in store.keys():
        raise KeyError(key)
    return delete_orig(key)


def _refuse_write(*args, **kwargs):
    raise RuntimeError("Write operation on read-only store.")


def _get_store(path):
    url = f"hfs://{path}"
    store = get_store_from_url(url)
    store.delete = partial(_check_and_delete, store=store, delete_orig=store.delete)
    return store


def _make_readonly(store):
    if callable(store):
        store = store()

    store.delete = _refuse_write
    store.put = _refuse_write

    return store


@pytest.fixture(scope="function")
def function_store(tmpdir):
    return partial(_get_store, tmpdir.join("function_store").strpath)


@pytest.fixture(scope="function")
def function_store_ro(function_store):
    return partial(_make_readonly, function_store)


@pytest.fixture(scope="function", params=["ro", "rw"])
def function_store_rwro(request, function_store, function_store_ro):
    if request.param == "ro":
        return function_store_ro
    else:
        return function_store


@pytest.fixture(scope="function")
def function_store2(tmpdir):
    return partial(_get_store, tmpdir.join("function_store2").strpath)


@pytest.fixture
def azure_store_cfg_factory(caplog):
    def _f(suffix):
        account_name = "qe72nmjyy6mbi"
        env_name = f"AZURE_BLOB_KEY_{account_name.upper()}"
        account_key = os.environ.get(env_name)
        if not account_key:
            pytest.skip(f"{env_name} not provided")

        container = "tests-{component}-{py_major}-{py_minor}-{suffix}".format(
            component="plateau",
            py_major=sys.version_info.major,
            py_minor=sys.version_info.minor,
            suffix=suffix,
        )

        return {
            "type": "hazure",
            "create_if_missing": True,
            "account_name": account_name,
            "account_key": account_key,
            "container": container,
        }

    return _f


@pytest.fixture(scope="module")
def module_store(tmpdir_factory):
    return partial(_get_store, tmpdir_factory.mktemp("module_store").realpath())


@pytest.fixture
def mock_uuid(mocker):
    uuid = mocker.patch("plateau.core.uuid._uuid_hook_str")
    uuid.return_value = "auto_dataset_uuid"
    return uuid


@pytest.fixture
def mock_default_metadata_version(mocker, backend_identifier):
    mock_metadata_version = 1

    # Mock `plateau.core.utils.verify_metadata_version`
    def patched__verify_metadata_version(metadata_version):
        pass

    mocker.patch(
        "plateau.core.utils._verify_metadata_version",
        patched__verify_metadata_version,
    )

    # Mock `plateau.io_components.metapartition.parse_input_to_metapartition`
    def patched__parse_input_to_metapartition(
        obj, table_name, metadata_version=None, *args, **kwargs
    ):
        if metadata_version == mock_metadata_version:
            return MetaPartition(
                label=gen_uuid(),
                data=obj,
                metadata_version=metadata_version,
                table_name=table_name,
            )
        try:
            return parse_input_to_metapartition(
                obj, table_name, metadata_version, *args, **kwargs
            )
        except ValueError as e:
            # Raise a "custom" error to distinguish this error from the error raised
            # by `parse_input_to_metapartition` when the object has not previously
            # passed through this mock function
            raise AssertionError(f"Traversed through mock. Original error: {e}") from e

    mocker.patch(
        f"plateau.io.{backend_identifier}.parse_input_to_metapartition",
        patched__parse_input_to_metapartition,
    )

    return mock_metadata_version


@pytest.fixture(params=[4], scope="session")
def metadata_version(request):
    if request.param < 4:
        with pytest.raises(ValueError):
            yield request.param
    else:
        yield request.param


def pytest_runtest_setup(item):
    # If a test is marked with `@pytest.mark.min_metadata_version(x)` the test will
    # be skipped if the `metadata_version` fixture is below this value
    if "metadata_version" in getattr(item, "fixturenames", []):
        minimal_version = item.get_closest_marker("min_metadata_version")
        if minimal_version:
            minimal_version = minimal_version.args[0]
            if minimal_version > item.callspec.params["metadata_version"]:
                pytest.skip("Skipped since the metadata version is too low")


@pytest.fixture(params=["json", "msgpack"], scope="session")
def metadata_storage_format(request):
    return request.param


@pytest.fixture
def df_all_types_schema(df_all_types):
    return make_meta(df_all_types, origin="df_all_types")


@pytest.fixture
def df_all_types_empty_schema(df_all_types):
    df_empty = df_all_types.drop(0)
    assert df_empty.empty
    return make_meta(df_empty, origin="df_empty")


@pytest.fixture
def df_all_types():
    return get_dataframe_alltypes()


@pytest.fixture
def df_not_nested():
    return get_dataframe_not_nested()


def _get_meta_partitions_with_dataframe(metadata_version):
    df = pd.DataFrame(
        OrderedDict(
            [
                ("P", [1]),
                ("L", [1]),
                ("TARGET", [1]),
                ("DATE", [datetime.date(2010, 1, 1)]),
            ]
        )
    )
    mp = MetaPartition(label="cluster_1", data=df, metadata_version=metadata_version)
    df_2 = pd.DataFrame(
        OrderedDict(
            [
                ("P", [2]),
                ("L", [2]),
                ("TARGET", [2]),
                ("DATE", [datetime.date(2009, 12, 31)]),
            ]
        )
    )
    mp2 = MetaPartition(label="cluster_2", data=df_2, metadata_version=metadata_version)
    return [mp, mp2]


@pytest.fixture(scope="session")
def meta_partitions_dataframe(metadata_version):
    """Create a list of MetaPartitions for testing.

    The partitions include in-memory pd.DataFrames without external
    references, i.e. files  are empty
    """
    with cm_frozen_time(TIME_TO_FREEZE):
        return _get_meta_partitions_with_dataframe(metadata_version)


@pytest.fixture(scope="function")
def meta_partitions_dataframe_function(metadata_version):
    """Create a list of MetaPartitions for testing.

    The partitions include in-memory pd.DataFrames without external
    references, i.e. files  are empty
    """
    return _get_meta_partitions_with_dataframe(metadata_version)


@pytest.fixture(scope="session")
def meta_partitions_evaluation_dataframe(metadata_version):
    """Create a list of MetaPartitions for testing.

    The partitions include in-memory pd.DataFrames without external
    references, i.e. files  are empty
    """
    df = pd.DataFrame(
        OrderedDict([("P", [1]), ("L", [1]), ("HORIZON", [1]), ("PRED", [10])])
    )
    mp = MetaPartition(label="cluster_1_1", data=df, metadata_version=metadata_version)
    df_2 = pd.DataFrame(
        OrderedDict([("P", [1]), ("L", [1]), ("HORIZON", [2]), ("PRED", [20])])
    )
    mp2 = MetaPartition(
        label="cluster_1_2", data=df_2, metadata_version=metadata_version
    )
    df_3 = pd.DataFrame(
        OrderedDict([("P", [2]), ("L", [2]), ("HORIZON", [1]), ("PRED", [10])])
    )
    mp3 = MetaPartition(
        label="cluster_2_1", data=df_3, metadata_version=metadata_version
    )
    df_4 = pd.DataFrame(
        OrderedDict([("P", [2]), ("L", [2]), ("HORIZON", [2]), ("PRED", [20])])
    )
    mp4 = MetaPartition(
        label="cluster_2_2", data=df_4, metadata_version=metadata_version
    )
    return [mp, mp2, mp3, mp4]


def _store_metapartitions(meta_partitions_dataframe, store):
    result = []
    for mp in meta_partitions_dataframe:
        mp = mp.store_dataframes(store=store, dataset_uuid="dataset_uuid")
        result.append(mp)
    return result


@pytest.fixture(scope="session")
def meta_partitions_complete(meta_partitions_dataframe, store_session):
    return _store_metapartitions(meta_partitions_dataframe, store_session)


@pytest.fixture(scope="function")
def meta_partitions_complete_function(meta_partitions_dataframe_function, store):
    return _store_metapartitions(meta_partitions_dataframe_function, store)


@pytest.fixture(scope="session")
def meta_partitions_files_only(meta_partitions_complete):
    """Create a list of MetaPartitions for testing.

    The partitions only include external file references with data
    stored in store
    """
    result = []
    for mp in meta_partitions_complete:
        mp = mp.copy(data={})
        result.append(mp)
    return result


@pytest.fixture(scope="function")
def meta_partitions_files_only_function(meta_partitions_complete_function):
    """Create a list of MetaPartitions for testing.

    The partitions only include external file references with data
    stored in store
    """
    result = []
    for mp in meta_partitions_complete_function:
        mp = mp.copy(data={})
        result.append(mp)
    return result


@pytest.fixture(scope="session")
def meta_partitions_evaluation_files_only(
    meta_partitions_evaluation_dataframe, store_session
):
    """Create a list of MetaPartitions for testing.

    The partitions only include external file references with data
    stored in store
    """
    result = []
    for mp in meta_partitions_evaluation_dataframe:
        mp = mp.store_dataframes(store=store_session, dataset_uuid="evaluation_uuid")
        mp = mp.copy(data={})
        result.append(mp)
    return result


@pytest.fixture(scope="session")
def dataset(meta_partitions_files_only, store_session_factory):
    """Create a proper plateau dataset in store with two partitions."""
    with cm_frozen_time(TIME_TO_FREEZE):
        return store_dataset_from_partitions(
            partition_list=meta_partitions_files_only,
            dataset_uuid="dataset_uuid",
            store=store_session_factory(),
            dataset_metadata={"dataset": "metadata"},
        )


@pytest.fixture(scope="session")
def dataset_factory(dataset, store_session_factory):
    return DatasetFactory(
        dataset_uuid=dataset.uuid,
        store_factory=store_session_factory,
        load_schema=True,
        load_all_indices=False,
    )


@pytest.fixture(scope="session")
def dataset_partition_keys(meta_partitions_dataframe, store_session_factory):
    """Create a proper plateau dataset in store with two partitions."""
    with cm_frozen_time(TIME_TO_FREEZE):
        new_mps = []
        for mp in meta_partitions_dataframe:
            new_mps.append(mp.partition_on(["P"]))
        new_mps = _store_metapartitions(new_mps, store_session_factory())

        return store_dataset_from_partitions(
            partition_list=new_mps,
            dataset_uuid="dataset_uuid_partition_keys",
            store=store_session_factory(),
            dataset_metadata={"dataset": "metadata"},
        )


@pytest.fixture(scope="session")
def dataset_partition_keys_factory(dataset_partition_keys, store_session_factory):
    return DatasetFactory(
        dataset_uuid=dataset_partition_keys.uuid,
        store_factory=store_session_factory,
        load_schema=True,
        load_all_indices=False,
    )


@pytest.fixture(scope="session")
def dataset_with_index(
    meta_partitions_dataframe, store_session_factory, metadata_version
):
    """Create a proper plateau dataset in store with two partitions."""
    with cm_frozen_time(TIME_TO_FREEZE):
        new_mps = []
        for mp in meta_partitions_dataframe:
            mp = mp.partition_on(["P"])
            mp = mp.build_indices(["L"])
            new_mps.append(mp)

        new_mps = _store_metapartitions(new_mps, store_session_factory())
        return store_dataset_from_partitions(
            partition_list=new_mps,
            dataset_uuid="dataset_uuid_partition_keys_index",
            store=store_session_factory(),
            dataset_metadata={"dataset": "metadata"},
        )


@pytest.fixture(scope="session")
def dataset_with_index_factory(dataset_with_index, store_session_factory):
    return DatasetFactory(
        dataset_uuid=dataset_with_index.uuid,
        store_factory=store_session_factory,
        load_schema=True,
        load_all_indices=False,
    )


@pytest.fixture(scope="function")
def dataset_function(meta_partitions_files_only_function, store):
    """Create a proper plateau dataset in store with two partitions."""
    with cm_frozen_time(TIME_TO_FREEZE):
        return store_dataset_from_partitions(
            partition_list=meta_partitions_files_only_function,
            dataset_uuid="dataset_uuid",
            store=store,
            dataset_metadata={"dataset": "metadata"},
        )


@pytest.fixture(scope="session")
def evaluation_dataset(meta_partitions_evaluation_files_only, store_session):
    with cm_frozen_time(TIME_TO_FREEZE):
        return store_dataset_from_partitions(
            partition_list=meta_partitions_evaluation_files_only,
            dataset_uuid="evaluation_uuid",
            store=store_session,
        )



from .utils import create_dataset


def test_garbage_collect_idempotent(store_factory, garbage_collect_callable):
    """Check that garbage collection does nothing when there is no garbage."""

    create_dataset("uuid", store_factory, 4)

    keys_before = set(store_factory().keys())
    garbage_collect_callable("uuid", store_factory)
    keys_after = set(store_factory().keys())
    assert keys_before == keys_after


def _test_gc(uuid, store_factory, garbage_collect_callable):
    store = store_factory()

    keys_before = set(store.keys())

    # Add a non-tracked table file
    store.put(f"{uuid}/table/trash.parquet", b"trash")

    # Add a non-tracked index file
    store.put(f"{uuid}/indices/trash.parquet", b"trash")

    garbage_collect_callable(uuid, store_factory)

    keys_after = set(store.keys())
    assert keys_before == keys_after


def test_gc_tables(store_factory, garbage_collect_callable):
    create_dataset("uuid", store_factory, 4)
    _test_gc("uuid", store_factory, garbage_collect_callable)


def test_gc_without_secondary_indices(
    store_factory, garbage_collect_callable, dataset_function
):
    _test_gc("dataset_uuid", store_factory, garbage_collect_callable)



import pickle
from functools import partial

import dask.bag as db
import pytest

from plateau.io.dask.bag import store_bag_as_dataset
from plateau.io.dask.delayed import store_delayed_as_dataset
from plateau.io.testing.write import *  # noqa


def _store_dataframes(execution_mode, df_list, *args, **kwargs):
    if execution_mode == "dask.bag":
        bag = store_bag_as_dataset(db.from_sequence(df_list), *args, **kwargs)

        s = pickle.dumps(bag, pickle.HIGHEST_PROTOCOL)
        bag = pickle.loads(s)

        return bag.compute()
    elif execution_mode == "dask.delayed":
        tasks = store_delayed_as_dataset(df_list, *args, **kwargs)

        s = pickle.dumps(tasks, pickle.HIGHEST_PROTOCOL)
        tasks = pickle.loads(s)

        return tasks.compute()
    else:
        raise ValueError(f"Unknown execution mode: {execution_mode}")


@pytest.fixture(params=["dask.delayed", "dask.bag"])
def bound_store_dataframes(request):
    return partial(_store_dataframes, request.param)



from functools import partial

import pytest
from minimalkv import KeyValueStore, get_store_from_url
from minimalkv.decorator import PrefixDecorator

from plateau.core.utils import ensure_store, lazy_store


class NoPickle:
    def __getstate__(self):
        raise RuntimeError("do NOT pickle this object!")


def mark_nopickle(obj):
    obj._nopickle = NoPickle()


def no_pickle_store(url):
    store = get_store_from_url(url)
    mark_nopickle(store)
    return store


def no_pickle_factory(url):
    return partial(no_pickle_store, url)


@pytest.fixture(params=["URL", "KeyValue", "Callable"])
def store_input_types(request, tmpdir):
    url = f"hfs://{tmpdir}"

    if request.param == "URL":
        return url
    elif request.param == "KeyValue":
        return get_store_from_url(url)
    elif request.param == "Callable":
        return no_pickle_factory(url)
    else:
        raise RuntimeError(f"Encountered unknown store type {type(request.param)}")


def test_ensure_store(store_input_types):
    store = ensure_store(store_input_types)
    assert isinstance(store, KeyValueStore)
    value = b"value"
    key = "key"
    store.put(key, value)
    assert value == store.get(key)

    assert store is ensure_store(store)


def test_ensure_store_fact(store_input_types):
    store_fact = lazy_store(store_input_types)
    assert callable(store_fact)
    store = store_fact()
    assert isinstance(store, KeyValueStore)
    value = b"value"
    key = "key"
    store.put(key, value)
    assert value == store.get(key)

    assert store_fact is lazy_store(store_fact)


def test_ensure_store_returns_same_store():
    store = get_store_from_url("memory://")
    assert ensure_store(lambda: store) is store


def test_lazy_store_returns_same_store():
    store = get_store_from_url("memory://")
    assert lazy_store(lambda: store)() is store


def test_lazy_store_accepts_decorated_store():
    store = get_store_from_url("memory://")
    pstore = PrefixDecorator("pre", store)
    assert lazy_store(pstore)() is pstore



#!/usr/bin/env python
import os

import pyarrow as pa
from minimalkv import get_store_from_url

from plateau.core.testing import get_dataframe_alltypes
from plateau.serialization import ParquetSerializer

if __name__ == "__main__":
    ser = ParquetSerializer()
    dir_path = os.path.dirname(os.path.realpath(__file__))
    store = get_store_from_url(f"hfs://{dir_path}")

    df = get_dataframe_alltypes()
    df["byte"] = b"\x82\xd6\xc1\x06Z\x08\x11\xe9\x85eJ\x00\x07\xf8\n\x10"
    ref_file = f"{pa.__version__}"
    ser.store(store, ref_file, df)



from urllib.parse import quote as quote_python

from plateau.core.urlencode import (
    quote as quote_ktk,
    unquote as unquote_ktk,
)

TEST_STRING = "Test string with lots of special characters !@#$%^&*+=()[]\\{}<>?|'\"-_~`☺✌☕file.par"


def test_urlquoting_backwards_compatibility():
    """This tests asserts that unquoting strings encoded with urlquote does
    produce the same result as unquoting strings quoted with pythons urllib
    (the encodings differ for Python < 3.7)."""
    assert unquote_ktk(quote_ktk(TEST_STRING)) == unquote_ktk(quote_python(TEST_STRING))



from plateau.serialization._generic import filter_array_like, filter_df_from_predicates
from plateau.serialization.testing import get_dataframe_not_nested


class TimeFilterDF:
    params = ["conjunctions", "disjunctions"]

    def setup(self, predicate):
        if predicate == "conjunctions":
            self.predicate = [
                [
                    ("int16", ">", 123),
                    ("int32", "<", 321),
                    ("bool", "==", True),
                    ("bool", "==", True),
                ]
            ]
        elif predicate == "disjunctions":
            self.predicate = [
                [("int16", ">", 123)],
                [("int32", "<", 321)],
                [("int32", "<", 321)],
                [("int32", "<", 321)],
            ]
        self.df = get_dataframe_not_nested(10**5)

    def time_filter_df_from_predicates(self, predicate):
        filter_df_from_predicates(self.df, self.predicate)


cols_to_filter = [
    # All numpy native types perform in the same order as integers therefore it
    # should be sufficient to test those (including datetime)
    "int64",
    # Object/strings are the true cost drivers and should be included in the benchmark
    "unicode",
]


class TimeFilterArray:
    params = cols_to_filter

    def setup(self, column):
        if column == "null":
            raise NotImplementedError()
        self.arr = (
            get_dataframe_not_nested(10**5)
            .sample(frac=1.0)
            .reset_index(drop=True)[column]
            .values
        )
        self.value = self.arr[12345]

    def time_eq(self, column):
        filter_array_like(self.arr, "==", self.value)

    def time_ge(self, column):
        filter_array_like(self.arr, ">=", self.value)

    def time_le(self, column):
        filter_array_like(self.arr, "<=", self.value)


class TimeFilterArrayIn:
    params = (
        cols_to_filter,
        [10, 100, 1000],
        [10**4, 10**5, 10**6],
    )
    param_names = ["column", "filter_size", "array_size", "enabled"]

    def setup(self, column, filter_size, array_size):
        if column == "null":
            raise NotImplementedError()
        ser = (
            get_dataframe_not_nested(array_size)
            .sample(frac=1.0)
            .reset_index(drop=True)[column]
        )
        self.arr = ser.values
        self.value = ser.sample(n=filter_size).unique()

    def time_in(self, column, filter_size, array_size):
        filter_array_like(self.arr, "in", self.value)


if __name__ == "__main__":
    restore = TimeFilterArrayIn()
    restore.setup("bool", 100, 100, -1)
    restore.time_in("bool", 100, 100, -1)



import numpy as np
import numpy.testing as npt
import pandas as pd
import pandas.testing as pdt
import pytest

from plateau.utils.pandas import (
    aggregate_to_lists,
    concat_dataframes,
    drop_sorted_duplicates_keep_last,
    is_dataframe_sorted,
    mask_sorted_duplicates_keep_last,
    merge_dataframes_robust,
    sort_dataframe,
)


class TestConcatDataframes:
    @pytest.fixture(params=[True, False])
    def dummy_default(self, request):
        if request.param:
            return pd.DataFrame(data={"a": [-2, -3], "b": 1.0}, columns=["a", "b"])
        else:
            return None

    @pytest.fixture(params=[True, False])
    def maybe_iter(self, request):
        if request.param:
            return iter
        else:
            return list

    def test_many(self, dummy_default, maybe_iter):
        dfs = [
            pd.DataFrame(
                data={"a": [0, 1], "b": 1.0}, columns=["a", "b"], index=[10, 11]
            ),
            pd.DataFrame(
                data={"a": [2, 3], "b": 2.0}, columns=["a", "b"], index=[10, 11]
            ),
            pd.DataFrame(data={"a": [4, 5], "b": 3.0}, columns=["a", "b"]),
        ]
        expected = pd.DataFrame(
            {"a": [0, 1, 2, 3, 4, 5], "b": [1.0, 1.0, 2.0, 2.0, 3.0, 3.0]},
            columns=["a", "b"],
        )

        actual = concat_dataframes(maybe_iter(dfs), dummy_default)
        pdt.assert_frame_equal(actual, expected)

    def test_single(self, dummy_default, maybe_iter):
        df = pd.DataFrame(
            data={"a": [0, 1], "b": 1.0}, columns=["a", "b"], index=[10, 11]
        )

        actual = concat_dataframes(maybe_iter([df.copy()]), dummy_default)
        pdt.assert_frame_equal(actual, df)

    def test_default(self, maybe_iter):
        df = pd.DataFrame(
            data={"a": [0, 1], "b": 1.0}, columns=["a", "b"], index=[10, 11]
        )

        actual = concat_dataframes(maybe_iter([]), df)
        pdt.assert_frame_equal(actual, df)

    def test_fail_no_default(self, maybe_iter):
        with pytest.raises(ValueError) as exc:
            concat_dataframes(maybe_iter([]), None)
        assert str(exc.value) == "Cannot concatenate 0 dataframes."

    @pytest.mark.parametrize(
        "dfs",
        [
            [pd.DataFrame({"a": [0, 1]})],
            [pd.DataFrame({"a": [0, 1]}), pd.DataFrame({"a": [2, 3]})],
        ],
    )
    def test_whipe_list(self, dfs):
        concat_dataframes(dfs)
        assert dfs == []

    @pytest.mark.parametrize(
        "dfs,expected",
        [
            (
                # dfs
                [pd.DataFrame(index=range(3))],
                # expected
                pd.DataFrame(index=range(3)),
            ),
            (
                # dfs
                [pd.DataFrame(index=range(3)), pd.DataFrame(index=range(2))],
                # expected
                pd.DataFrame(index=range(5)),
            ),
        ],
    )
    def test_no_columns(self, dfs, expected):
        actual = concat_dataframes(dfs)
        pdt.assert_frame_equal(actual, expected)

    def test_fail_different_colsets(self, maybe_iter):
        dfs = [pd.DataFrame({"a": [1]}), pd.DataFrame({"a": [1], "b": [2]})]
        with pytest.raises(
            ValueError, match="Not all DataFrames have the same set of columns!"
        ):
            concat_dataframes(maybe_iter(dfs))


@pytest.mark.parametrize(
    "df,columns",
    [
        (
            # df
            pd.DataFrame({"a": [1, 2, 3], "b": [1, 2, 3]}),
            # columns
            ["a"],
        ),
        (
            # df
            pd.DataFrame({"a": [3, 2, 1], "b": [1, 2, 3]}),
            # columns
            ["a"],
        ),
        (
            # df
            pd.DataFrame({"a": [3, 2, 1, 3, 2, 1], "b": [2, 2, 2, 1, 1, 1]}),
            # columns
            ["a", "b"],
        ),
        (
            # df
            pd.DataFrame({"a": [3, 2, 1], "b": [1, 2, 3]}, index=[1000, 2000, 3000]),
            # columns
            ["a"],
        ),
        (
            # df
            pd.DataFrame({"a": [3.0, 2.0, 1.0], "b": [1, 2, 3]}),
            # columns
            ["a"],
        ),
        (
            # df
            pd.DataFrame({"a": ["3", "2", "1"], "b": [1, 2, 3]}),
            # columns
            ["a"],
        ),
        (
            # df
            pd.DataFrame({"a": [True, False], "b": [1, 2]}),
            # columns
            ["a"],
        ),
        (
            # df
            pd.DataFrame(
                {
                    "a": [
                        pd.Timestamp("2018-01-03"),
                        pd.Timestamp("2018-01-02"),
                        pd.Timestamp("2018-01-01"),
                    ],
                    "b": [1, 2, 3],
                }
            ),
            # columns
            ["a"],
        ),
        (
            # df
            pd.DataFrame(
                {"a": pd.Series(["3", "2", "1"]).astype("category"), "b": [1, 2, 3]}
            ),
            # columns
            ["a"],
        ),
    ],
)
def test_sort_dataframe(df, columns):
    expected = df.sort_values(columns).reset_index(drop=True)
    actual = sort_dataframe(df, columns)
    pdt.assert_frame_equal(actual, expected)


@pytest.mark.parametrize(
    "df,columns,expected_mask",
    [
        (
            # df
            pd.DataFrame({"a": [1, 2, 3]}),
            # columns
            ["a"],
            # expected_mask
            np.array([False, False, False]),
        ),
        (
            # df
            pd.DataFrame({"a": [1, 1, 3]}),
            # columns
            ["a"],
            # expected_mask
            np.array([True, False, False]),
        ),
        (
            # df
            pd.DataFrame({"a": [1, 1, 3], "b": [1, 2, 3]}),
            # columns
            ["a"],
            # expected_mask
            np.array([True, False, False]),
        ),
        (
            # df
            pd.DataFrame({"a": [1, 1, 3], "b": [1, 2, 3]}),
            # columns
            ["a", "b"],
            # expected_mask
            np.array([False, False, False]),
        ),
        (
            # df
            pd.DataFrame({"a": [1, 1, 3], "b": [1, 1, 3]}),
            # columns
            ["a", "b"],
            # expected_mask
            np.array([True, False, False]),
        ),
        (
            # df
            pd.DataFrame({"a": [1]}),
            # columns
            ["a"],
            # expected_mask
            np.array([False]),
        ),
        (
            # df
            pd.DataFrame({"a": []}),
            # columns
            ["a"],
            # expected_mask
            np.array([], dtype=bool),
        ),
        (
            # df
            pd.DataFrame(
                {
                    "a": [1, 1, 3],
                    "b": [1.0, 1.0, 3.0],
                    "c": ["a", "a", "b"],
                    "d": [True, True, False],
                    "e": [
                        pd.Timestamp("2018"),
                        pd.Timestamp("2018"),
                        pd.Timestamp("2019"),
                    ],
                    "f": pd.Series(["a", "a", "b"]).astype("category"),
                }
            ),
            # columns
            ["a", "b", "c", "d", "e", "f"],
            # expected_mask
            np.array([True, False, False]),
        ),
        (
            # df
            pd.DataFrame({"a": [1, 2, 3, 4, 4, 5, 6, 6]}),
            # columns
            ["a"],
            # expected_mask
            np.array([False, False, False, True, False, False, True, False]),
        ),
        (
            # df
            pd.DataFrame({"a": [2, 2, 3]}),
            # columns
            [],
            # expected_mask
            np.array([False, False, False]),
        ),
        (
            # df
            pd.DataFrame({"a": [1, 1, 3]}, index=[1000, 2000, 1]),
            # columns
            ["a"],
            # expected_mask
            np.array([True, False, False]),
        ),
    ],
)
def test_sorted_duplicates_keep_last(df, columns, expected_mask):
    actual_mask = mask_sorted_duplicates_keep_last(df, columns)
    assert actual_mask.dtype == bool
    npt.assert_array_equal(actual_mask, expected_mask)

    actual_df = drop_sorted_duplicates_keep_last(df, columns)
    expected_df = df[~expected_mask]
    pdt.assert_frame_equal(actual_df, expected_df)

    if columns:
        # pandas crashes for empty column lists
        pd_mask = df.duplicated(subset=columns, keep="last").values
        pd_df = df.drop_duplicates(subset=columns, keep="last")
        npt.assert_array_equal(pd_mask, expected_mask)
        pdt.assert_frame_equal(pd_df, expected_df)


@pytest.mark.parametrize(
    "df_input,by",
    [
        (
            # df_input
            pd.DataFrame({"x": [0], "y": [0], "v": ["a"]}),
            # by
            ["x", "y"],
        ),
        (
            # df_input
            pd.DataFrame({"x": [0, 0], "y": [0, 0], "v": ["a", "b"]}),
            # by
            ["x", "y"],
        ),
        (
            # df_input
            pd.DataFrame({"x": [0, 0], "y": [0, 0], "v": ["a", "a"]}),
            # by
            ["x", "y"],
        ),
        (
            # df_input
            pd.DataFrame(
                {
                    "x": [1, 0, 0, 1, 1],
                    "y": [1, 0, 0, 0, 1],
                    "v": ["a", "b", "c", "d", "e"],
                }
            ),
            # by
            ["x", "y"],
        ),
        (
            # df_input
            pd.DataFrame({"x": [], "y": [], "v": []}),
            # by
            ["x", "y"],
        ),
        (
            # df_input
            pd.DataFrame({"x": [0, 0], "y": [0, 0], "v": ["a", "a"]}),
            # by
            [],
        ),
    ],
)
def test_aggregate_to_lists(df_input, by):
    data_col = "v"

    # pandas is broken for empty DFs
    if df_input.empty:
        df_expected = df_input
    else:
        if by:
            df_expected = df_input.groupby(by=by, as_index=False, observed=True)[
                data_col
            ].agg(lambda series: list(series.values))
        else:
            df_expected = pd.DataFrame(
                {data_col: pd.Series([list(df_input[data_col].values)])}
            )

    df_actual = aggregate_to_lists(df_input, by, data_col)

    pdt.assert_frame_equal(df_actual, df_expected)


def test_is_dataframe_sorted_no_cols():
    df = pd.DataFrame({})
    with pytest.raises(ValueError, match="`columns` must contain at least 1 column"):
        is_dataframe_sorted(df, [])


@pytest.mark.parametrize(
    "df,columns",
    [
        (
            # df
            pd.DataFrame({"x": []}),
            # columns
            ["x"],
        ),
        (
            # df
            pd.DataFrame({"x": [], "y": [], "z": []}),
            # columns
            ["x", "y", "z"],
        ),
        (
            # df
            pd.DataFrame({"x": [0, 1, 10]}),
            # columns
            ["x"],
        ),
        (
            # df
            pd.DataFrame({"x": [0, 1, 10], "y": [20, 21, 210]}),
            # columns
            ["x", "y"],
        ),
        (
            # df
            pd.DataFrame({"x": [0, 1, 1], "y": [10, 0, 1]}),
            # columns
            ["x", "y"],
        ),
        (
            # df
            pd.DataFrame({"x": [0, 1], "y": [1, 0]}),
            # columns
            ["x"],
        ),
        (
            # df
            pd.DataFrame(
                {
                    "x": [0, 0, 0, 0, 1, 1, 1, 1],
                    "y": [0, 0, 1, 1, 0, 0, 1, 1],
                    "z": [0, 1, 0, 1, 0, 1, 0, 1],
                }
            ),
            # columns
            ["x", "y", "z"],
        ),
        (
            # df
            pd.DataFrame({"x": [0, 0], "y": [0, 0], "z": [0, 0]}),
            # columns
            ["x", "y", "z"],
        ),
        (
            # df
            pd.DataFrame({"x": pd.Series(["1", "2"]).astype("category")}),
            # columns
            ["x"],
        ),
    ],
)
def test_assert_df_sorted_ok(df, columns):
    assert is_dataframe_sorted(df, columns)


@pytest.mark.parametrize(
    "df,columns",
    [
        (
            # df
            pd.DataFrame({"x": [1, 0]}),
            # columns
            ["x"],
        ),
        (
            # df
            pd.DataFrame({"x": [0, 1, 1], "y": [0, 1, 0]}),
            # columns
            ["x", "y"],
        ),
        (
            # df
            pd.DataFrame({"x": [0, 0], "y": [0, 0], "z": [1, 0]}),
            # columns
            ["x", "y", "z"],
        ),
        (
            # df
            pd.DataFrame({"x": [0, 0], "y": [1, 0], "z": [0, 0]}),
            # columns
            ["x", "y", "z"],
        ),
        (
            # df
            pd.DataFrame({"x": [1, 0], "y": [0, 0], "z": [0, 0]}),
            # columns
            ["x", "y", "z"],
        ),
        (
            # df
            pd.DataFrame({"x": pd.Series(["2", "1"]).astype("category")}),
            # columns
            ["x"],
        ),
    ],
)
def test_assert_df_sorted_no(df, columns):
    assert not is_dataframe_sorted(df, columns)


@pytest.mark.parametrize(
    "df1,df2,how,expected",
    [
        (
            # df1
            pd.DataFrame({"i": [0, 1], "x": [0, 1], "v1": [11, 12]}),
            # df2
            pd.DataFrame({"x": [0, 1], "v2": [21, 22]}),
            # how
            "inner",
            # expected
            pd.DataFrame({"i": [0, 1], "x": [0, 1], "v1": [11, 12], "v2": [21, 22]}),
        ),
        (
            # df1
            pd.DataFrame({"i": [0, 1], "x": [0, 1], "v1": [11, 12]}),
            # df2
            pd.DataFrame({"x": [0, 1], "v2": [21, 22]}),
            # how
            "left",
            # expected
            pd.DataFrame({"i": [0, 1], "x": [0, 1], "v1": [11, 12], "v2": [21, 22]}),
        ),
        (
            # df1
            pd.DataFrame({"i": [0, 1], "x": [0, 1], "v1": [11, 12]}),
            # df2
            pd.DataFrame({"x": [0], "v2": [21]}),
            # how
            "inner",
            # expected
            pd.DataFrame({"i": [0], "x": [0], "v1": [11], "v2": [21]}),
        ),
        (
            # df1
            pd.DataFrame({"i": [0, 1], "x": [0, 1], "v1": [11, 12]}),
            # df2
            pd.DataFrame({"x": [0], "v2": [21]}),
            # how
            "left",
            # expected
            pd.DataFrame(
                {"i": [0, 1], "x": [0, 1], "v1": [11, 12], "v2": [21, np.nan]}
            ),
        ),
        (
            # df1
            pd.DataFrame({"i": [0, 1], "v1": [11, 12]}),
            # df2
            pd.DataFrame({"v2": [21, 22]}),
            # how
            "inner",
            # expected
            pd.DataFrame(
                {"i": [0, 0, 1, 1], "v1": [11, 11, 12, 12], "v2": [21, 22, 21, 22]}
            ),
        ),
        (
            # df1
            pd.DataFrame({"i": [0, 1], "v1": [11, 12]}),
            # df2
            pd.DataFrame({"v2": [21, 22]}),
            # how
            "left",
            # expected
            pd.DataFrame(
                {"i": [0, 0, 1, 1], "v1": [11, 11, 12, 12], "v2": [21, 22, 21, 22]}
            ),
        ),
        (
            # df1
            pd.DataFrame({"i": [0, 1], "v1": [11, 12]}),
            # df2
            pd.DataFrame({"v2": pd.Series([], dtype=int)}),
            # how
            "inner",
            # expected
            pd.DataFrame(
                {
                    "i": pd.Series([], dtype=int),
                    "v1": pd.Series([], dtype=int),
                    "v2": pd.Series([], dtype=int),
                }
            ),
        ),
        (
            # df1
            pd.DataFrame({"i": [0, 1], "v1": [11, 12]}),
            # df2
            pd.DataFrame({"v2": pd.Series([], dtype=int)}),
            # how
            "left",
            # expected
            pd.DataFrame({"i": [0, 1], "v1": [11, 12], "v2": [np.nan, np.nan]}),
        ),
    ],
)
def test_merge_dataframes_robust(df1, df2, how, expected):
    df1_backup = df1.copy()
    df2_backup = df2.copy()

    actual = merge_dataframes_robust(df1, df2, how)
    actual = actual.sort_values(sorted(actual.columns)).reset_index(drop=True)

    pdt.assert_frame_equal(df1, df1_backup)
    pdt.assert_frame_equal(df2, df2_backup)

    pdt.assert_frame_equal(actual, expected)



"""Workarounds for limitations of the minimalkv API."""

import logging
import time
from urllib.parse import quote

from minimalkv.contrib import VALID_KEY_RE_EXTENDED

try:
    # azure-storage-blob < 12
    from azure.common import (
        AzureMissingResourceHttpError as _AzureMissingResourceHttpError,
    )
    from azure.storage.blob import BlockBlobService as _BlockBlobService
except ImportError:

    class _BlockBlobService:  # type: ignore
        """Dummy class."""

    class _AzureMissingResourceHttpError:  # type: ignore
        """Dummy class."""


try:
    # azure-storage-blob >= 12
    from azure.core.exceptions import ResourceNotFoundError as _ResourceNotFoundError
    from azure.storage.blob import ContainerClient as _ContainerClient
except ImportError:

    class _ContainerClient:  # type: ignore
        """Dummy class."""

    class _ResourceNotFoundError:  # type: ignore
        """Dummy class."""


__all__ = ("copy_keys",)


_logger = logging.getLogger(__name__)


# Specialized implementation for azure-storage-blob < 12, using BlockBlobService (`bbs`):


def _has_azure_bbs(store):
    try:
        # store decorators will forward getattr calls
        return isinstance(store.block_blob_service, _BlockBlobService)
    except AttributeError:
        return False


def _azure_bbs_content_md5(block_blob_service, container, key, accept_missing=False):
    try:
        return block_blob_service.get_blob_properties(
            container, key
        ).properties.content_settings.content_md5
    except _AzureMissingResourceHttpError:
        if accept_missing:
            return None
        else:
            raise KeyError(key) from None


def _copy_azure_bbs(keys, src_store, tgt_store):
    src_container = src_store.container
    tgt_container = tgt_store.container
    src_bbs = src_store.block_blob_service
    tgt_bbs = tgt_store.block_blob_service

    cprops = {}
    for k in keys:
        source_md5 = _azure_bbs_content_md5(
            src_bbs, src_container, k, accept_missing=False
        )

        if source_md5 is None:
            _logger.debug(f"Missing hash for {k}")
        else:
            tgt_md5 = _azure_bbs_content_md5(
                tgt_bbs, tgt_container, k, accept_missing=True
            )

            if source_md5 == tgt_md5:
                _logger.debug(f"Omitting copy to {k} (checksum match)")
                continue

        copy_source = src_bbs.make_blob_url(
            src_container, quote(k), sas_token=src_bbs.sas_token
        )
        cprops[k] = tgt_bbs.copy_blob(tgt_container, k, copy_source)

    for k, cprop in cprops.items():
        while True:
            blob = tgt_bbs.get_blob_properties(tgt_container, k)
            cprop_current = blob.properties.copy
            assert cprop.id == cprop_current.id, f"Concurrent copy to {k}"
            if cprop_current.status == "pending":
                _logger.debug(f"Waiting for pending copy to {k}...")
                time.sleep(0.1)
                continue
            elif cprop_current.status == "success":
                _logger.debug(f"Copy to {k} completed")
                break  # break from while, continue in for-loop
            else:
                raise RuntimeError(
                    f"Error while copying: status is {cprop_current.status}: {cprop_current.status_description}"
                )


# Specialized implementation for azure-storage-blob >= 12, using ContainerClient (`cc`):
def _has_azure_cc(store):
    try:
        # store decorators will forward getattr calls
        return isinstance(store.blob_container_client, _ContainerClient)
    except AttributeError:
        return False


def _azure_cc_content_md5(cc, key, accept_missing=False):
    try:
        bc = cc.get_blob_client(key)
        return bc.get_blob_properties().content_settings.content_md5
    except _ResourceNotFoundError:
        if accept_missing:
            return None
        else:
            raise KeyError(key) from None


def _copy_azure_cc(keys, src_store, tgt_store):
    src_cc = src_store.blob_container_client
    tgt_cc = tgt_store.blob_container_client

    copy_ids = {}
    for k in keys:
        source_md5 = _azure_cc_content_md5(src_cc, k, accept_missing=False)

        if source_md5 is None:
            _logger.debug(f"Missing hash for {k}")
        else:
            tgt_md5 = _azure_cc_content_md5(tgt_cc, k, accept_missing=True)

            if source_md5 == tgt_md5:
                _logger.debug(f"Omitting copy to {k} (checksum match)")
                continue

        copy_source = src_cc.get_blob_client(k).url
        copy_ids[k] = tgt_cc.get_blob_client(k).start_copy_from_url(copy_source)[
            "copy_id"
        ]

    for k, copy_id in copy_ids.items():
        while True:
            cprop_current = tgt_cc.get_blob_client(k).get_blob_properties().copy
            assert copy_id == cprop_current.id, f"Concurrent copy to {k}"
            if cprop_current.status == "pending":
                _logger.debug(f"Waiting for pending copy to {k}...")
                time.sleep(0.1)
                continue
            elif cprop_current.status == "success":
                _logger.debug(f"Copy to {k} completed")
                break  # break from while, continue in for-loop
            else:
                raise RuntimeError(
                    f"Error while copying: status is {cprop_current.status}: {cprop_current.status_description}"
                )


def _copy_naive(keys, src_store, tgt_store):
    for k in keys:
        tgt_store.put(k, src_store.get(k))


def copy_keys(keys, src_store, tgt_store):
    """Copy keys from one store the another.

    Parameters
    ----------
    keys: Iterable[str]
        Keys to copy.
    src_store: Union[minimalkv.KeyValueStore, Callable[[], minimalkv.KeyValueStore]]
        Source KV store.
    tgt_store: Union[minimalkv.KeyValueStore, Callable[[], minimalkv.KeyValueStore]]
        Target KV store.
    """
    if callable(src_store):
        src_store = src_store()
    if callable(tgt_store):
        tgt_store = tgt_store()

    keys = sorted(keys)
    for k in keys:
        if (k is None) or (not VALID_KEY_RE_EXTENDED.match(k)) or (k == "/"):
            raise ValueError(f"Illegal key: {k}")

    if _has_azure_bbs(src_store) and _has_azure_bbs(tgt_store):
        _logger.debug(
            "Azure stores based on BlockBlobStorage class detected, use fast-path."
        )
        _copy_azure_bbs(keys, src_store, tgt_store)
    elif _has_azure_cc(src_store) and _has_azure_cc(tgt_store):
        _logger.debug(
            "Azure stores based on ContainerClient class detected, use fast-path."
        )
        _copy_azure_cc(keys, src_store, tgt_store)
    else:
        _logger.debug("Use naive slow-path.")
        _copy_naive(keys, src_store, tgt_store)



import pandas as pd
import pytest

from plateau.io.dask.dataframe import collect_dataset_metadata
from plateau.io.eager import store_dataframes_as_dataset, update_dataset_from_dataframes
from plateau.io_components.metapartition import _METADATA_SCHEMA
from plateau.serialization import ParquetSerializer


def test_collect_dataset_metadata(store_session_factory, dataset):
    df_stats = collect_dataset_metadata(
        store=store_session_factory,
        dataset_uuid="dataset_uuid",
        predicates=None,
        frac=1,
    ).compute()

    actual = df_stats.drop(
        columns=[
            "row_group_compressed_size",
            "row_group_uncompressed_size",
            "serialized_size",
        ],
    )
    actual.sort_values(by=["partition_label", "row_group_id"], inplace=True)

    expected = pd.DataFrame(
        data={
            "partition_label": ["cluster_1", "cluster_2"],
            "row_group_id": [0, 0],
            "number_rows_total": [1, 1],
            "number_row_groups": [1, 1],
            "number_rows_per_row_group": [1, 1],
        },
        index=[0, 0],
    )
    pd.testing.assert_frame_equal(actual, expected)


def test_collect_dataset_metadata_predicates(store_session_factory, dataset):
    predicates = [[("P", "==", 1)]]

    df_stats = collect_dataset_metadata(
        store=store_session_factory,
        dataset_uuid="dataset_uuid",
        predicates=predicates,
        frac=1,
    ).compute()

    actual = df_stats.drop(
        columns=[
            "row_group_compressed_size",
            "row_group_uncompressed_size",
            "serialized_size",
        ]
    )
    actual.sort_values(by=["partition_label", "row_group_id"], inplace=True)

    # Predicates are only evaluated on index level and have therefore no effect on this dataset
    expected = pd.DataFrame(
        data={
            "partition_label": ["cluster_1", "cluster_2"],
            "row_group_id": [0, 0],
            "number_rows_total": [1, 1],
            "number_row_groups": [1, 1],
            "number_rows_per_row_group": [1, 1],
        },
        index=[0, 0],
    )
    pd.testing.assert_frame_equal(actual, expected)


def test_collect_dataset_metadata_predicates_on_index(store_factory):
    df = pd.DataFrame(
        data={"P": range(10), "L": ["a", "a", "a", "a", "a", "b", "b", "b", "b", "b"]}
    )
    store_dataframes_as_dataset(
        store=store_factory,
        dataset_uuid="dataset_uuid",
        partition_on=["L"],
        dfs=[df],
    )
    predicates = [[("L", "==", "b")]]

    df_stats = collect_dataset_metadata(
        store=store_factory,
        dataset_uuid="dataset_uuid",
        predicates=predicates,
        frac=1,
    ).compute()

    assert "L=b" in df_stats["partition_label"].values[0]

    df_stats.sort_values(by=["partition_label", "row_group_id"], inplace=True)
    actual = df_stats.drop(
        columns=[
            "partition_label",
            "row_group_compressed_size",
            "row_group_uncompressed_size",
            "serialized_size",
        ],
    )

    expected = pd.DataFrame(
        data={
            "row_group_id": [0],
            "number_rows_total": [5],
            "number_row_groups": [1],
            "number_rows_per_row_group": [5],
        },
    )
    pd.testing.assert_frame_equal(actual, expected)


def test_collect_dataset_metadata_predicates_row_group_size(store_factory):
    ps = ParquetSerializer(chunk_size=2)
    df = pd.DataFrame(
        data={"P": range(10), "L": ["a", "a", "a", "a", "a", "b", "b", "b", "b", "b"]}
    )
    store_dataframes_as_dataset(
        store=store_factory,
        dataset_uuid="dataset_uuid",
        partition_on=["L"],
        dfs=[df],
        df_serializer=ps,
    )

    predicates = [[("L", "==", "a")]]

    df_stats = collect_dataset_metadata(
        store=store_factory,
        dataset_uuid="dataset_uuid",
        predicates=predicates,
        frac=1,
    ).compute()

    for part_label in df_stats["partition_label"]:
        assert "L=a" in part_label
    df_stats.sort_values(by=["partition_label", "row_group_id"], inplace=True)

    actual = df_stats.drop(
        columns=[
            "partition_label",
            "row_group_compressed_size",
            "row_group_uncompressed_size",
            "serialized_size",
        ],
    )

    expected = pd.DataFrame(
        data={
            "row_group_id": [0, 1, 2],
            "number_rows_total": [5, 5, 5],
            "number_row_groups": [3, 3, 3],
            "number_rows_per_row_group": [2, 2, 1],
        },
        index=[0, 1, 2],
    )
    pd.testing.assert_frame_equal(actual, expected)


def test_collect_dataset_metadata_frac_smoke(store_session_factory, dataset):
    df_stats = collect_dataset_metadata(
        store=store_session_factory,
        dataset_uuid="dataset_uuid",
        frac=0.8,
    ).compute()
    columns = {
        "partition_label",
        "row_group_id",
        "row_group_compressed_size",
        "row_group_uncompressed_size",
        "number_rows_total",
        "number_row_groups",
        "serialized_size",
        "number_rows_per_row_group",
    }

    assert set(df_stats.columns) == columns


def test_collect_dataset_metadata_empty_dataset(store_factory):
    df = pd.DataFrame(columns=["A", "b"], index=pd.RangeIndex(start=0, stop=0))
    store_dataframes_as_dataset(
        store=store_factory, dataset_uuid="dataset_uuid", dfs=[df], partition_on=["A"]
    )
    df_stats = collect_dataset_metadata(
        store=store_factory, dataset_uuid="dataset_uuid"
    ).compute()
    expected = pd.DataFrame(columns=_METADATA_SCHEMA.keys())
    expected = expected.astype(_METADATA_SCHEMA)
    pd.testing.assert_frame_equal(expected, df_stats)


def test_collect_dataset_metadata_concat(store_factory):
    """Smoke-test concatenation of empty and non-empty dataset metadata
    collections."""
    df = pd.DataFrame(data={"A": [1, 1, 1, 1], "b": [1, 1, 2, 2]})
    store_dataframes_as_dataset(
        store=store_factory, dataset_uuid="dataset_uuid", dfs=[df], partition_on=["A"]
    )
    df_stats1 = collect_dataset_metadata(
        store=store_factory,
        dataset_uuid="dataset_uuid",
    ).compute()

    # Remove all partitions of the dataset
    update_dataset_from_dataframes(
        [], store=store_factory, dataset_uuid="dataset_uuid", delete_scope=[{"A": 1}]
    )

    df_stats2 = collect_dataset_metadata(
        store=store_factory,
        dataset_uuid="dataset_uuid",
    ).compute()
    pd.concat([df_stats1, df_stats2])


def test_collect_dataset_metadata_delete_dataset(store_factory):
    df = pd.DataFrame(data={"A": [1, 1, 1, 1], "b": [1, 1, 2, 2]})
    store_dataframes_as_dataset(
        store=store_factory, dataset_uuid="dataset_uuid", dfs=[df], partition_on=["A"]
    )
    # Remove all partitions of the dataset
    update_dataset_from_dataframes(
        [], store=store_factory, dataset_uuid="dataset_uuid", delete_scope=[{"A": 1}]
    )

    df_stats = collect_dataset_metadata(
        store=store_factory,
        dataset_uuid="dataset_uuid",
    ).compute()
    expected = pd.DataFrame(columns=_METADATA_SCHEMA)
    expected = expected.astype(_METADATA_SCHEMA)
    pd.testing.assert_frame_equal(expected, df_stats)


def test_collect_dataset_metadata_fraction_precision(store_factory):
    df = pd.DataFrame(data={"A": range(100), "B": range(100)})

    store_dataframes_as_dataset(
        store=store_factory,
        dataset_uuid="dataset_uuid",
        dfs=[df],
        partition_on=["A"],
    )  # Creates 100 partitions

    df_stats = collect_dataset_metadata(
        store=store_factory, dataset_uuid="dataset_uuid", frac=0.2
    ).compute()
    assert len(df_stats) == 20


def test_collect_dataset_metadata_at_least_one_partition(store_factory):
    """Make sure we return at least one partition, even if none would be
    returned by rounding frac * n_partitions."""
    df = pd.DataFrame(data={"A": range(100), "B": range(100)})

    store_dataframes_as_dataset(
        store=store_factory,
        dataset_uuid="dataset_uuid",
        dfs=[df],
        partition_on=["A"],
    )  # Creates 100 partitions

    df_stats = collect_dataset_metadata(
        store=store_factory, dataset_uuid="dataset_uuid", frac=0.005
    ).compute()
    assert len(df_stats) == 1


def test_collect_dataset_metadata_invalid_frac(store_session_factory, dataset):
    with pytest.raises(ValueError, match="Invalid value for parameter `frac`"):
        collect_dataset_metadata(
            store=store_session_factory,
            dataset_uuid="dataset_uuid",
            frac=1.1,
        )

    with pytest.raises(ValueError, match="Invalid value for parameter `frac`"):
        collect_dataset_metadata(
            store=store_session_factory,
            dataset_uuid="dataset_uuid",
            frac=0.0,
        )



from plateau.core.common_metadata import *  # noqa: F401, F403
from plateau.core.dataset import *  # noqa: F401, F403
from plateau.core.factory import *  # noqa: F401, F403
from plateau.core.index import *  # noqa: F401, F403
from plateau.core.typing import *  # noqa: F401, F403
from plateau.core.utils import *  # noqa: F401, F403



"""Code was taken from pandas.

https://github.com/pandas-dev/pandas/blob/c5a47118d73a53a016a28103726b57e27924bf29/pandas/util/_decorators.py#L66

with the following license:

BSD 3-Clause License

Copyright (c) 2008-2012, AQR Capital Management, LLC, Lambda Foundry, Inc. and PyData Development Team
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

* Redistributions of source code must retain the above copyright notice, this
  list of conditions and the following disclaimer.

* Redistributions in binary form must reproduce the above copyright notice,
  this list of conditions and the following disclaimer in the documentation
  and/or other materials provided with the distribution.

* Neither the name of the copyright holder nor the names of its
  contributors may be used to endorse or promote products derived from
  this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

import warnings
from functools import wraps


def deprecate_kwarg(old_arg_name, new_arg_name, mapping=None, stacklevel=2):
    """Decorator to deprecate a keyword argument of a function.

    Parameters
    ----------
    old_arg_name : str
        Name of argument in function to deprecate
    new_arg_name : str or None
        Name of preferred argument in function. Use None to raise warning that
        ``old_arg_name`` keyword is deprecated.
    mapping : dict or callable
        If mapping is present, use it to translate old arguments to
        new arguments. A callable must do its own value checking;
        values not found in a dict will be forwarded unchanged.
    Examples
    --------
    The following deprecates 'cols', using 'columns' instead
    >>> @deprecate_kwarg(old_arg_name='cols', new_arg_name='columns')
    ... def f(columns=''):
    ...     print(columns)
    ...
    >>> f(columns='should work ok')
    should work ok
    >>> f(cols='should raise warning')
    FutureWarning: cols is deprecated, use columns instead
      warnings.warn(msg, FutureWarning)
    should raise warning
    >>> f(cols='should error', columns="can\'t pass do both")
    TypeError: Can only specify 'cols' or 'columns', not both
    >>> @deprecate_kwarg('old', 'new', {'yes': True, 'no': False})
    ... def f(new=False):
    ...     print('yes!' if new else 'no!')
    ...
    >>> f(old='yes')
    FutureWarning: old='yes' is deprecated, use new=True instead
      warnings.warn(msg, FutureWarning)
    yes!
    To raise a warning that a keyword will be removed entirely in the future
    >>> @deprecate_kwarg(old_arg_name='cols', new_arg_name=None)
    ... def f(cols='', another_param=''):
    ...     print(cols)
    ...
    >>> f(cols='should raise warning')
    FutureWarning: the 'cols' keyword is deprecated and will be removed in a
    future version please takes steps to stop use of 'cols'
    should raise warning
    >>> f(another_param='should not raise warning')
    should not raise warning
    >>> f(cols='should raise warning', another_param='')
    FutureWarning: the 'cols' keyword is deprecated and will be removed in a
    future version please takes steps to stop use of 'cols'
    should raise warning
    """

    if mapping is not None and not hasattr(mapping, "get") and not callable(mapping):
        raise TypeError(
            "mapping from old to new argument values must be dict or callable!"
        )

    def _deprecate_kwarg(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            old_arg_value = kwargs.pop(old_arg_name, None)

            if new_arg_name is None and old_arg_value is not None:
                msg = (
                    f"the '{old_arg_name}' keyword is deprecated and will be "
                    "removed in a future version. "
                    f"Please take steps to stop the use of '{old_arg_name}'"
                )
                warnings.warn(msg, FutureWarning, stacklevel=stacklevel)
                kwargs[old_arg_name] = old_arg_value
                return func(*args, **kwargs)

            if old_arg_value is not None:
                if mapping is not None:
                    if hasattr(mapping, "get"):
                        new_arg_value = mapping.get(old_arg_value, old_arg_value)
                    else:
                        new_arg_value = mapping(old_arg_value)
                    msg = (
                        f"the {old_arg_name}={old_arg_value!r} keyword is deprecated, "
                        f"use {new_arg_name}={new_arg_value!r} instead"
                    )
                else:
                    new_arg_value = old_arg_value
                    msg = (
                        f"the '{old_arg_name}' keyword is deprecated, "
                        f"use '{new_arg_name}' instead"
                    )

                warnings.warn(msg, FutureWarning, stacklevel=stacklevel)
                if kwargs.get(new_arg_name, None) is not None:
                    msg = (
                        f"Can only specify '{old_arg_name}' or '{new_arg_name}', "
                        "not both"
                    )
                    raise TypeError(msg)
                else:
                    kwargs[new_arg_name] = new_arg_value
            return func(*args, **kwargs)

        return wrapper

    return _deprecate_kwarg



import logging
from collections.abc import Iterable
from copy import copy
from typing import Any, TypeVar, cast

import numpy as np
import pandas as pd
import pyarrow as pa
import pyarrow.parquet as pq
from toolz.itertoolz import partition_all

import plateau.core._time
from plateau.core import naming
from plateau.core._compat import ARROW_GE_20
from plateau.core._mixins import CopyMixin
from plateau.core.common_metadata import normalize_type
from plateau.core.docs import default_docs
from plateau.core.typing import StoreInput
from plateau.core.urlencode import quote
from plateau.core.utils import ensure_store
from plateau.serialization import (
    PredicatesType,
    check_predicates,
    filter_array_like,
    filter_df_from_predicates,
    filter_predicates_by_column,
)

ValueType = TypeVar("ValueType")
IndexDictType = dict[ValueType, list[str]]

_logger = logging.getLogger(__name__)

_PARTITION_COLUMN_NAME = "partition"

__all__ = (
    "IndexBase",
    "ExplicitSecondaryIndex",
    "PartitionIndex",
)


class IndexBase(CopyMixin):
    """Initialize an IndexBase.

    Parameters
    ----------
    column
        Name of the column this index is for.
    index_dct
        Mapping from index values to partition labels
    dtype
        Type of index. If left out and ``index_dct`` is present, this will be inferred.
    normalize_dtype
        Normalize type information and values within ``index_dct``. The user may
        disable this when it the index was already normalized, e.g. when the
        index python objects gets copied, or when the index data is restored
        from a parquet file that was written by a trusted write path.
    """

    def __init__(
        self,
        column: str,
        index_dct: IndexDictType | None = None,
        dtype: pa.DataType | None = None,
        normalize_dtype: bool = True,
    ):
        if column == _PARTITION_COLUMN_NAME:
            raise ValueError(
                f"Cannot create an index for column {column} due to an internal implementation conflict. "
                "Please contact a plateau maintainer if you receive this error message"
            )
        if (dtype is None) and index_dct:
            # do dtype given but index_dct is present => auto-derive dtype
            table = _index_dct_to_table(index_dct, column, None)
            schema = table.schema
            dtype = schema[0].type

        if dtype is not None:
            if pa.types.is_nested(dtype):
                raise NotImplementedError("Indices w/ nested types are not supported")
            if pa.types.is_null(dtype):
                raise NotImplementedError("Indices w/ null/NA type are not supported")
            if normalize_dtype:
                dtype, _t_pd, _t_np, _metadata = normalize_type(dtype, None, None, None)

        self.column = column
        self.dtype = dtype
        self.creation_time = plateau.core._time.datetime_utcnow()
        self.index_dct: IndexDictType

        self.index_dct = {}
        self._index_dct_available = False
        if index_dct is not None:
            self._index_dct_available = True
            if normalize_dtype:
                # index_dct may be from an untrusted or weakly typed source, so we need to normalize and fuse values
                self.index_dct = {}
                n_collisions = 0
                for value, partitions in index_dct.items():
                    value = IndexBase.normalize_value(self.dtype, value)
                    if value not in self.index_dct:
                        self.index_dct[value] = copy(partitions)
                    else:
                        existing = self.index_dct[value]
                        self.index_dct[value] += [
                            part for part in partitions if part not in existing
                        ]
                        n_collisions += 1

                if n_collisions:
                    _logger.warning(
                        f"Value normalization for index column {column} resulted in {n_collisions} collision(s). plateau merged "
                        "the affected partition lists, but you may want to check if this was desired."
                    )
            else:
                # data comes from a trusted source (e.g. an index that we've already preserved and are now reading), so
                # use the fast path and don't re-normalize the values
                self.index_dct = index_dct

        super().__init__()

    def copy(self, **kwargs) -> "IndexBase":
        return super().copy(normalize_dtype=False, **kwargs)

    def __repr__(self) -> str:
        repr_str = []
        for key, val in self.__dict__.items():
            if isinstance(val, dict):
                repr_str.append("=".join([key, str(sorted(val.keys()))]))
            else:
                repr_str.append("=".join([key, str(val)]))
        return "{class_}({attrs})".format(
            class_=type(self).__name__, attrs=", ".join(repr_str)
        )

    def observed_values(
        self, date_as_object=True, coerce_temporal_nanoseconds=True
    ) -> np.ndarray:
        """Return an array of all observed values."""
        keys = np.array(list(self.index_dct.keys()))
        labeled_array = _safe_paarray(keys, self.dtype)

        _coerce = {"coerce_temporal_nanoseconds": coerce_temporal_nanoseconds}
        return np.array(
            labeled_array.to_pandas(date_as_object=date_as_object, **_coerce)
        )

    @staticmethod
    def normalize_value(dtype: pa.DataType, value: Any) -> Any:
        """Normalize value according to index dtype.

        This may apply casts (e.g. integers to floats) or parsing (e.g. timestamps from strings) to the value.

        Parameters
        ----------
        dtype
            Arrow type of the index.
        value
            any value

        Returns
        -------
        value: Any
            normalized value, with a type that matches the index dtype

        Raises
        ------
        ValueError
            If dtype of the index was not set or derived.
        NotImplementedError
            If the dtype cannot be handled.
        """
        if dtype is None:
            raise ValueError(
                "Cannot normalize index values as long as dtype is not set"
            )
        elif pa.types.is_string(dtype) or pa.types.is_large_string(dtype):
            if isinstance(value, bytes):
                return value.decode("utf-8")
            else:
                return str(value)
        elif pa.types.is_binary(dtype):
            if isinstance(value, bytes):
                return value
            else:
                return str(value).encode("utf-8")
        elif pa.types.is_date(dtype):
            return pd.Timestamp(value).date()
        elif pa.types.is_temporal(dtype):
            return pd.Timestamp(value).to_datetime64().astype("datetime64[ns]")
        elif pa.types.is_integer(dtype):
            return int(value)
        elif pa.types.is_floating(dtype):
            return float(value)
        elif pa.types.is_boolean(dtype):
            if isinstance(value, str):
                if value.lower() == "false":
                    return False
                elif value.lower() == "true":
                    return True
                else:
                    return ValueError(f'Cannot parse boolean value "{value}"')
            return bool(value)
        else:
            raise NotImplementedError(f"Cannot normalize index value for type {dtype}")

    @property
    def loaded(self) -> bool:
        """Check if the index was already loaded into memory."""
        return self._index_dct_available

    def eval_operator(self, op: str, value: ValueType) -> set[str]:
        """Evaluates a given operator on the index for a given value and
        returns all partition labels allowed by this index.

        Parameters
        ----------
        op: str
            A string representation of the operator to be evaluated. Supported are
            "==", "<=", ">=", "<", ">", "in"
            For details, see documentation of plateau.serialization
        value: object
            The value to be evaluated
        Returns
        -------
        set
            Allowed partition labels
        """
        result = set()
        index_dct = self.index_dct
        index_type = self.dtype
        index_arr = None
        if index_type is not None and index_type and not pa.types.is_date(index_type):
            index_type = index_type.to_pandas_dtype()
            try:
                index_arr = np.fromiter(index_dct.keys(), dtype=index_type)
            except ValueError:
                pass
        if index_arr is None:
            index_arr = np.array(list(index_dct.keys()))

        index = filter_array_like(
            index_arr, op, value, strict_date_types=True, column_name=self.column
        )
        allowed_values = index_arr[index]
        # Need to determine allowed values to include predicates like `in`
        for value in allowed_values:
            result.update(set(self.index_dct[value]))
        return result

    def query(self, value: ValueType) -> list[str]:
        """Query this index for a given value. Raises an exception if the index
        is external and not loaded.

        Parameters
        ----------
        value:
            The value that is looked up in the index dictionary.

        Returns
        -------
        keys: List[str]
            A list of keys of partitions that contain the corresponding value.
        """
        if self.index_dct is None:
            raise RuntimeError(
                "Index is external. To query an external index, the index ne to be preloaded."
            )

        return self.index_dct.get(IndexBase.normalize_value(self.dtype, value), [])

    def to_dict(self) -> IndexDictType:
        """Serialise the object to Python object that can be part of a larger
        dictionary that may be serialised to JSON."""
        if self.index_dct is None:
            raise RuntimeError("Index dict not set.")
        return self.index_dct

    def update(self, index: "IndexBase", inplace: bool = False) -> "IndexBase":
        """Returns a new Index object in case of a change.

        The new index object will no longer carry the attribute `index_storage_key`
        since it is no longer a proper representation of the stored index object.

        Parameters
        ----------
        index: [plateau.core.index.IndexBase]
            The index which should be added to this one
        """
        if not isinstance(index, IndexBase):
            raise TypeError(
                f"Need to input an plateau.core.index.IndexBase object, instead got `{type(index)}`"
            )
        # Assume that if one of the indices dtypes is None, they are compatible. In future versions we should make
        # dtype a non-optional parameter
        if self.dtype is not None and index.dtype is not None:
            if self.dtype != index.dtype:
                raise TypeError(
                    f"Trying to update an index with different types. Expected `{self.dtype}` but got `{index.dtype}`"
                )

        if self.column != index.column:
            raise ValueError(
                f"Trying to update an index with the wrong column. Got `{index.column}` but expected `{self.column}`"
            )

        if index.index_dct is None or len(index.index_dct) == 0:
            return self
        if inplace:
            new_index_dict = self.index_dct
        else:
            new_index_dict = copy(self.index_dct)

        for value, partition_list in index.index_dct.items():
            old = new_index_dict.get(value, [])
            new_index_dict[value] = list(set(old + partition_list))
        return self.copy(column=self.column, index_dct=new_index_dict, dtype=self.dtype)

    def remove_partitions(
        self, list_of_partitions: list[str], inplace: bool = False
    ) -> "IndexBase":
        """Removes a partition from the internal index dictionary.

        The new index object will no longer carry the attribute `index_storage_key`
        since it is no longer a proper representation of the stored index object.

        Parameters
        ----------
        list_of_partitions
            The partition to be removed
        inplace
            If `True` the operation is performed inplace and will return the same object
        """
        if not list_of_partitions:
            return self
        partitions_to_delete = set(list_of_partitions)
        if inplace:
            values_to_remove = set()
            for val, partition_list in self.index_dct.items():
                new_partition_set = set(partition_list) - partitions_to_delete
                if new_partition_set:
                    self.index_dct[val][:] = new_partition_set
                else:
                    values_to_remove.add(val)
            for val in values_to_remove:
                del self.index_dct[val]
            # Call the constructor again to reinit the creation timestamp
            return self.copy(
                column=self.column, index_dct=self.index_dct, dtype=self.dtype
            )
        else:
            new_index_dict = {}
            for val, partition_list in self.index_dct.items():
                new_partition_set = set(partition_list) - partitions_to_delete
                if new_partition_set:
                    new_index_dict[val] = list(new_partition_set)
            return self.copy(
                column=self.column, index_dct=new_index_dict, dtype=self.dtype
            )

    def remove_values(
        self, list_of_values: list[str], inplace: bool = False
    ) -> "IndexBase":
        """Removes a value from the internal index dictionary.

        Parameters
        ----------
        list_of_values
            The value to be removed
        inplace:
            If `True` the operation is performed inplace and will return the same object
        """
        set_of_values = {
            IndexBase.normalize_value(self.dtype, v) for v in list_of_values
        }

        if not set_of_values:
            return self

        if inplace:
            for value in set_of_values:
                if value in self.index_dct:
                    del self.index_dct[value]
            # new object for new timestamp, etc
            return self.copy(
                column=self.column,
                index_dct=self.index_dct,
                dtype=self.dtype,
                normalize_dtype=False,
            )
        else:
            new_index_dict = {}
            for val, partition_list in self.index_dct.items():
                if val not in set_of_values:
                    new_index_dict[val] = partition_list

            return self.copy(
                column=self.column,
                index_dct=new_index_dict,
                dtype=self.dtype,
                normalize_dtype=False,
            )

    def __eq__(self, other) -> bool:
        if not isinstance(other, IndexBase):
            return False
        if self.column != other.column:
            return False
        if (self.dtype is None) and (other.dtype is not None):
            return False
        if (self.dtype is not None) and (other.dtype is None):
            return False
        if self.dtype != other.dtype:
            return False
        if len(self.index_dct) != len(other.index_dct):
            return False
        for col, partition_list in self.index_dct.items():
            if col not in other.index_dct:
                return False
            if set(partition_list) != set(other.index_dct[col]):
                return False
        return True

    def __ne__(self, other) -> bool:
        return not (self == other)

    @default_docs
    def as_flat_series(
        self,
        compact: bool = False,
        partitions_as_index: bool = False,
        date_as_object: bool = False,
        predicates: PredicatesType = None,
    ) -> pd.Series:
        """Convert the Index object to a pandas.Series.

        Parameters
        ----------
        compact
            If True, ensures that the index will be unique. If there a multiple partition values per index, there values
            will be compacted into a list (see Examples section).
        partitions_as_index
            If True, the relation between index values and partitions will be reverted for the output dataframe:
            partition values will be used as index and the indices will be mapped to the partitions.
        predicates
            A list of predicates. If a literal within the provided predicates
            references a column which is not part of this index, this literal is
            interpreted as True.

        Examples
        --------
        >>> import pyarrow as pa
        >>> from plateau.core.index import ExplicitSecondaryIndex
        >>> index1 = ExplicitSecondaryIndex(
        ...     column="col", index_dct={1: ["part_1", "part_2"]}, dtype=pa.int64()
        ... )
        >>> index1.as_flat_series()
        col
        1    part_1
        1    part_2
        Name: partition, dtype: object
        >>> index1.as_flat_series(compact=True)
        col
        1    [part_1, part_2]
        Name: partition, dtype: object
        >>> index1.as_flat_series(partitions_as_index=True)
        partition
        part_1    1
        part_2    1
        Name: col, dtype: int64
        """
        check_predicates(predicates)
        table = _index_dct_to_table(
            self.index_dct, column=self.column, dtype=self.dtype
        )
        _coerce = {"coerce_temporal_nanoseconds": True}
        df = table.to_pandas(date_as_object=date_as_object, **_coerce)

        if predicates is not None:
            # If there is a conjunction without any reference to the index
            # column the entire predicates expression is evaluated to True. In
            # this case we do not need to filter the dataframe anymore
            for conjunction in predicates:
                new_conjunction = filter_predicates_by_column(
                    [conjunction], [self.column]
                )
                if new_conjunction is None:
                    break
            else:
                filtered_predicates = filter_predicates_by_column(
                    predicates, [self.column]
                )
                df = filter_df_from_predicates(df, predicates=filtered_predicates)

        result_column = _PARTITION_COLUMN_NAME
        # This is the way the dictionary is directly translated
        # value: [partition]
        if compact and not partitions_as_index:
            return df.set_index(self.column)[result_column]

        # In all other circumstances we need a flat series first
        # value: part_1
        # value: part_2
        # value2: part_1
        if partitions_as_index or not compact:
            if len(df) == 0:
                keys = np.array([], dtype=df[_PARTITION_COLUMN_NAME].values.dtype)
            else:
                keys = np.concatenate(df[_PARTITION_COLUMN_NAME].values)

            lengths = df[_PARTITION_COLUMN_NAME].apply(len).values
            lengths = lengths.astype(int)
            values_index = np.repeat(np.arange(len(df)), lengths)
            values = df[self.column].values[values_index]

            df = pd.DataFrame({_PARTITION_COLUMN_NAME: keys, self.column: values})

        # if it is not inverted and not compact, we're done
        if partitions_as_index:
            result_index = _PARTITION_COLUMN_NAME
            if compact:
                df = df.groupby(df[result_index], observed=True).apply(
                    lambda x: x[self.column].tolist()
                )
                df.name = self.column
            else:
                df = df.set_index(result_index)[self.column]
        else:
            df = df.set_index(self.column)[_PARTITION_COLUMN_NAME]
        return df


class PartitionIndex(IndexBase):
    """An Index class representing partition indices (sometimes also referred
    to as primary indices). A PartitionIndex is usually constructed by parsing
    the partition filenames which encode index information.

    The constructor for this class should usually not be called explicitly but indices should be created by e.g.
    :meth:`plateau.core.dataset.DatasetMetadataBase.load_partition_indices`
    """

    def __init__(
        self,
        column: str,
        index_dct: IndexDictType | None = None,
        dtype: pa.DataType = None,
        normalize_dtype: bool = True,
    ):
        if dtype is None:
            raise ValueError(
                f'PartitionIndex dtype of column "{column}" cannot be None!'
            )
        super().__init__(
            column=column,
            index_dct=index_dct,
            dtype=dtype,
            normalize_dtype=normalize_dtype,
        )

    def __eq__(self, other):
        if not isinstance(other, PartitionIndex):
            return False
        return super().__eq__(other)


class ExplicitSecondaryIndex(IndexBase):
    """An Index class representing an explicit, secondary index which is
    calculated and stored next to the dataset.

    In contrast to the `PartitionIndex` this needs to be calculated by
    an explicit pass over the data. All mutations of this class will
    erase the reference to the physical file and the storage of the
    mutated object will write to a new storage key.
    """

    def __init__(
        self,
        column: str,
        index_dct: IndexDictType | None = None,
        index_storage_key: str | None = None,
        dtype: pa.DataType | None = None,
        normalize_dtype: bool = True,
    ):
        if (index_dct is None) and not index_storage_key:
            raise ValueError("No valid index source specified")
        if not index_storage_key and not index_dct and dtype is None:
            raise ValueError("Trying to create non-typesafe index")
        self.index_storage_key = index_storage_key
        super().__init__(
            column=column,
            index_dct=index_dct,
            dtype=dtype,
            normalize_dtype=normalize_dtype,
        )

    def copy(self, **kwargs) -> "ExplicitSecondaryIndex":
        if kwargs:
            index_storage_key = kwargs.pop("index_storage_key", None)
        else:
            index_storage_key = self.index_storage_key
        return cast(
            ExplicitSecondaryIndex,
            super(IndexBase, self).copy(index_storage_key=index_storage_key, **kwargs),
        )

    def unload(self) -> "IndexBase":
        """Drop index data to safe memory."""
        idx = self.copy(index_dct={}, index_storage_key=self.index_storage_key)
        idx._index_dct_available = False
        return idx

    def __eq__(self, other) -> bool:
        if not isinstance(other, ExplicitSecondaryIndex):
            return False
        if self.index_storage_key != other.index_storage_key:
            return False
        try:
            return super().__eq__(other)
        except TypeError:  # an `index_dct == None`
            if self.index_dct != other.index_dct:
                return False
            return True

    @staticmethod
    def from_v2(column: str, dct_or_str: str | IndexDictType) -> "IndexBase":
        """Create an index instance from a version 2 Python structure.

        Parameters
        ----------
        column:
            Name of the column this index provides lookup for
        dct_or_str:
            Either the storage key of the external index or the index itself
            as a Python object structure.

        Returns
        -------
        index: [plateau.core.index.ExplicitSecondaryIndex]
        """
        if isinstance(dct_or_str, str):
            # External index
            return ExplicitSecondaryIndex(column=column, index_storage_key=dct_or_str)
        else:
            return ExplicitSecondaryIndex(column=column, index_dct=dct_or_str)

    def store(self, store: StoreInput, dataset_uuid: str) -> str:
        """Store the index as a parquet file.

        If compatible, the new keyname will be the name stored under the attribute `index_storage_key`.
        If this attribute is None, a new key will be generated of the format

            `{dataset_uuid}/indices/{column}/{timestamp}.by-dataset-index.parquet`

        where the timestamp is in nanosecond accuracy and is created upon Index object initialization

        Parameters
        ----------
        store:
        dataset_uuid:
        """
        storage_key = None
        store = ensure_store(store)

        if (
            self.index_storage_key is not None
            and dataset_uuid
            and dataset_uuid in self.index_storage_key
        ):
            storage_key = self.index_storage_key
        if storage_key is None:
            storage_key = f"{dataset_uuid}/indices/{quote(self.column)}/{quote(self.creation_time.isoformat())}{naming.EXTERNAL_INDEX_SUFFIX}"

        # The arrow representation of index_dct requires a large amount of memory because strings are duplicated and
        # flattened into the buffer. To avoid a high peak memory usage, split the index_dct into chunks and only convert
        # one chunk a time to arrow.
        parts_iter = partition_all(10_000, self.index_dct.items())

        # Get first table explicit because its schema is required for ParquetWriter.
        try:
            table = _index_dct_to_table(dict(next(parts_iter)), self.column, self.dtype)
        except StopIteration:
            # index_dct was empty, just pass it entirely
            table = _index_dct_to_table(self.index_dct, self.column, self.dtype)

        buf = pa.BufferOutputStream()
        with pq.ParquetWriter(buf, schema=table.schema) as writer:
            writer.write_table(table)
            del table

            for part in parts_iter:
                writer.write_table(
                    _index_dct_to_table(dict(part), self.column, self.dtype)
                )

        store.put(storage_key, buf.getvalue().to_pybytes())
        return storage_key

    def load(self, store: StoreInput):
        """Load an external index into memory. Returns a new index object that
        contains the index dictionary. Returns itself if the index is internal
        or an already loaded index.

        Parameters
        ----------
        store
            Object that implements the .get method for file/object loading.

        Returns
        -------
        index: [plateau.core.index.ExplicitSecondaryIndex]
        """
        if self.loaded:
            return self

        store = ensure_store(store)

        assert self.index_storage_key is not None
        index_buffer = store.get(self.index_storage_key)
        index_dct, column_type = _parquet_bytes_to_dict(self.column, index_buffer)

        return ExplicitSecondaryIndex(
            column=self.column,
            index_dct=index_dct,
            dtype=column_type,
            index_storage_key=self.index_storage_key,
            normalize_dtype=False,
        )

    def __getstate__(self):
        if not self.loaded:
            return (self.column, self.index_storage_key, self.dtype, None)

        table = _index_dct_to_table(self.index_dct, self.column, self.dtype)
        buf = pa.BufferOutputStream()
        pq.write_table(table, buf)
        parquet_bytes = buf.getvalue().to_pybytes()
        # Since `self.dtype` will be inferred by parquet bytes, do not return
        # this argument during serialization to avoid unnecessary memory consumption
        return (self.column, self.index_storage_key, None, parquet_bytes)

    def __setstate__(self, state):
        column, storage_key, column_type, parquet_bytes = state
        if parquet_bytes is None:
            index_dct = None
        else:
            index_dct, column_type = _parquet_bytes_to_dict(column, parquet_bytes)
        self.__init__(  # type: ignore
            column=column,
            index_dct=index_dct,
            index_storage_key=storage_key,
            dtype=column_type,
            normalize_dtype=False,
        )


_MULTI_COLUMN_INDEX_DCT_TYPE = dict[str, IndexBase]


def merge_indices(
    list_of_indices: list[_MULTI_COLUMN_INDEX_DCT_TYPE],
) -> _MULTI_COLUMN_INDEX_DCT_TYPE:
    """Merge a list of index dictionaries.

    Parameters
    ----------
    list_of_indices
        A list of tuples holding index information

        Format: [ (partition_label, index_dict) ]
    """
    final_indices: _MULTI_COLUMN_INDEX_DCT_TYPE = {}
    if not list_of_indices:
        return final_indices

    # shortcut can only be applied to Index object since the dict path needs the label in the passed tuple
    if len(list_of_indices) == 1:
        return list_of_indices[0]
    elif len(list_of_indices) > 2:
        first = merge_indices(list_of_indices[0::2])
        second = merge_indices(list_of_indices[1::2])
        return merge_indices([first, second])

    for indices in list_of_indices:
        for column, index in indices.items():
            if column in final_indices:
                final_indices[column] = final_indices[column].update(index)
            else:
                final_indices[column] = index

    return final_indices


def remove_partitions_from_indices(
    index_dict: _MULTI_COLUMN_INDEX_DCT_TYPE, partitions: list[str]
):
    """Remove a given list of partitions from a plateau index dictionary.

    Parameters
    ----------
    index_dict
        A dictionary holding plateau indices
    partitions
        A list of partition labels which should be removed form the index objects
    """
    new_index_dict = {}
    for column, index in index_dict.items():
        new_index_dict[column] = index.remove_partitions(partitions)
    return new_index_dict


def filter_indices(index_dict: _MULTI_COLUMN_INDEX_DCT_TYPE, partitions: Iterable[str]):
    """Filter a plateau index dictionary such that only the provided list of
    partitions is included in the index dictionary.

    All indices must be embedded!

    Parameters
    ----------
    index_dict:
        A dictionary holding plateau indices
    partition_list:
        A list of partition labels which are allowed in the output dictionary
    """
    index_types = {}
    types = {}
    temp_index_dct: dict[str, dict[Any, list[str]]] = {}
    for column, index in index_dict.items():
        temp_index_dct[column] = {}
        types[column] = index.dtype
        index_types[column] = type(index)
        for index_value, partition_list in index.to_dict().items():
            temp_index_dct[column][index_value] = [
                part_label for part_label in partition_list if part_label in partitions
            ]

    final_index_dict: _MULTI_COLUMN_INDEX_DCT_TYPE = {}
    for column, column_index in temp_index_dct.items():
        final_index_dict[column] = index_types[column](
            column=column, index_dct=column_index, dtype=types[column]
        )
    return final_index_dict


def _parquet_bytes_to_dict(column: str, index_buffer: bytes):
    reader = pa.BufferReader(index_buffer)
    # This can be done much more efficient but would take a lot more
    # time to implement so this will be only done on request.
    table = pq.read_table(reader)
    column_type = table.schema.field(column).type

    # `datetime.datetime` objects have a precision of up to microseconds only, so arrow
    # parses the type to `pa.timestamp("us")`. Since the
    # values are normalized to `numpy.datetime64[ns]` anyways, we do not care about this
    # and load the column type as `pa.timestamp("ns")`
    if column_type == pa.timestamp("us"):
        column_type = pa.timestamp("ns")

    _coerce = {"coerce_temporal_nanoseconds": True}
    df = table.to_pandas(**_coerce)

    index_dct = dict(
        zip(
            df[column].values,
            (list(x) for x in df[_PARTITION_COLUMN_NAME].values),
            strict=False,
        )
    )
    return index_dct, column_type


def _index_dct_to_table(index_dct: IndexDictType, column: str, dtype: pa.DataType):
    keys_it = index_dct.keys()

    # find possible type probe
    if len(keys_it) > 0:
        probe = next(iter(keys_it))
        has_probe = True
    else:
        probe = None
        has_probe = False

    # type inference
    if (dtype is None) and has_probe:
        if isinstance(probe, np.datetime64):
            dtype = pa.timestamp(
                "ns"
            )  # workaround pyarrow type inference bug (ARROW-2554)
        elif isinstance(probe, pd.Timestamp):
            dtype = pa.timestamp(
                "ns"
            )  # workaround pyarrow type inference bug (ARROW-2554)
        elif isinstance(probe, np.bool_ | bool):
            dtype = pa.bool_()

    # fix pyarrow input
    if dtype is None:
        keys: np.ndarray | list[Any] = np.asarray(list(keys_it))
    else:
        if pa.types.is_unsigned_integer(dtype):
            # numpy might create object ndarrays here, which pyarrow might (for some reason) convert fo floats
            keys = list(keys_it)
        elif (
            dtype == pa.timestamp("ns")
            and has_probe
            and isinstance(probe, pd.Timestamp)
        ):
            keys = np.asarray([d.to_datetime64() for d in keys_it])
        else:
            keys = np.asarray(list(keys_it))

    # TODO: Remove work-around
    # This is because of ARROW-1646:
    #   [Python] pyarrow.array cannot handle NumPy scalar types
    # Additional note: pyarrow.array is supposed to infer type automatically.
    # But the inferred type is not enough to hold np.uint64. Until this is fixed in
    # upstream Arrow, we have to retain the following line
    if not index_dct:
        # the np.array dtype will be double which arrow cannot convert to the target type, so use an empty list instead
        labeled_array = pa.array([], type=dtype)
    else:
        labeled_array = _safe_paarray(keys, dtype)

    partition_array = pa.array(list(index_dct.values()), type=pa.list_(pa.string()))

    return pa.Table.from_arrays(
        [labeled_array, partition_array], names=[column, _PARTITION_COLUMN_NAME]
    )


def _safe_paarray(arr: np.ndarray, dtype: pa.DataType) -> pa.Array:
    if dtype is not None and pa.types.is_large_string(dtype) and not ARROW_GE_20:
        return pa.array(iter(arr), type=dtype)
    else:
        return pa.array(arr, type=dtype)



import warnings
from functools import partial

import pandas as pd
from dask import delayed

try:
    from cytoolz import map
except ImportError:
    pass

CATEGORICAL_EFFICIENCY_WARN_LIMIT = 100000


def _get_data(mp, table=None):
    """Task to avoid serialization of lambdas."""
    return mp.data


def _cast_categorical_to_index_cat(df, categories):
    try:
        return df.astype(categories)
    except ValueError as verr:
        # Should be fixed by pandas>=0.24.0
        if "buffer source array is read-only" in str(verr):
            new_cols = {}
            for cat in categories:
                new_cols[cat] = df[cat].astype(df[cat].dtype.categories.dtype)
                new_cols[cat] = new_cols[cat].astype(categories[cat])
            return df.assign(**new_cols)
        raise


def _construct_categorical(column, dataset_metadata_factory):
    dataset_metadata = dataset_metadata_factory.load_index(column)
    values = dataset_metadata.indices[column].index_dct.keys()
    if len(values) > CATEGORICAL_EFFICIENCY_WARN_LIMIT:
        warnings.warn(
            f"Column {column} has {len(values)} distinct values, reading as categorical may increase memory consumption.",
            stacklevel=2,
        )
    return pd.api.types.CategoricalDtype(values, ordered=False)


def _maybe_get_categoricals_from_index(dataset_metadata_factory, categoricals):
    """In case a categorical is requested for a column we have an index on,
    construct the categorical from the index."""
    categoricals_from_index = {}
    if categoricals:
        for column in categoricals:
            if column in dataset_metadata_factory.indices:
                cat_dtype = _construct_categorical(column, dataset_metadata_factory)
                categoricals_from_index[column] = cat_dtype
    return categoricals_from_index


def map_delayed(func, mps, **kwargs):
    func = partial(func, **kwargs)
    delayed_func = delayed(func)
    return map(delayed_func, mps)



from collections.abc import Iterable

from urlquote import (
    quote as urlquote_quote,
    unquote as urlquote_unquote,
)
from urlquote.quoting import PYTHON_3_7_QUOTING


def quote(value):
    """Performs percent encoding on a sequence of bytes.

    if the given value is of string type, it will be encoded. If the
    value is neither of string type nor bytes type, it will be cast
    using the `str` constructor before being encoded in UTF-8.
    """
    return urlquote_quote(value, quoting=PYTHON_3_7_QUOTING).decode("utf-8")


def unquote(value):
    """Decodes a urlencoded string and performs necessary decoding depending on
    the used python version."""
    return urlquote_unquote(value).decode("utf-8")


def decode_key(
    key: str,
) -> tuple[str, str, list, str] | tuple[str, None, list, None]:
    """Split a given key into its plateau components
    `{dataset_uuid}/{table}/{key_indices}/{filename}`

    Example:
        `uuid/table/index_col=1/index_col=2/partition_label.parquet`

    Returns
    -------
    dataset_uuid: str
    table: str
    key_indices: list
        The already unquoted list of index pairs
    file_: str
        The file name
    """
    key_components = key.split("/")
    dataset_uuid = key_components[0]
    if len(key_components) < 3:
        return key, None, [], None
    table = key_components[1]
    file_ = key_components[-1]
    key_indices = unquote_indices(key_components[2:-1])
    return dataset_uuid, table, key_indices, file_


def quote_indices(indices: Iterable[tuple[str, str]]) -> list[str]:
    """Urlencode a list of column-value pairs and encode them as:

        `quote(column)=quote(value)`

    Parameters
    ----------
    indices
        A list of tuples where each list entry is (column, value)

    Returns
    -------
    List[str]
        List with urlencoded column=value strings
    """
    quoted_pairs = []
    for column, value in indices:
        quoted_pairs.append(f"{quote(column)}={quote(value)}")

    return quoted_pairs


def unquote_indices(index_strings: list[str]) -> list[tuple[str, str]]:
    """Take a list of encoded column-value strings and decode them to tuples.

    input: `quote(column)=quote(value)`
    output `(column, value)`

    Parameters
    ----------
    indices
        A list of tuples where each list entry is (column, value)

    Returns
    -------
    List[str]
        List with column value pairs
    """
    indices = []
    for index_string in index_strings:
        split_string = index_string.split("=")
        if len(split_string) == 2:
            column, value = split_string
            indices.append((unquote(column), unquote(value)))
    return indices



import timeit


class AsvBenchmarkConfig:
    """A mixin class to define default configurations for our benchmarks."""

    # Use a timer which measured Wall time (asv default measures CPU time)
    timer = timeit.default_timer



# Write the benchmarking functions here.
# See "Writing benchmarks" in the asv docs for more information.

import datetime

import numpy as np
import pandas as pd

from plateau.core.common_metadata import make_meta
from plateau.io_components.metapartition import MetaPartition

from .config import AsvBenchmarkConfig


class TimeMetaPartition(AsvBenchmarkConfig):
    params = (
        [10**5, 10**6],
        [
            (np.int64, 123456789),
            (str, "abcdefgh"),
            ("object", datetime.date(2018, 1, 1)),
        ],
    )

    def setup(self, num_rows, dtype):
        df = pd.DataFrame(
            {"primary_key": pd.Series(dtype[1], index=range(num_rows), dtype=dtype[0])}
        )
        self.schema = make_meta(df, partition_keys=["primary_key"], origin="df")
        self.df = df.drop("primary_key", axis=1)
        self.mp = MetaPartition(
            label=f"primary_key={dtype[0]}/base_label",
            metadata_version=4,
            schema=self.schema,
        )

    def time_reconstruct_index(self, num_rows, dtype):
        self.mp._reconstruct_index_columns(
            df=self.df,
            key_indices=[("primary_key", str(dtype[1]))],
            columns=None,
            categories=None,
            date_as_object=False,
        )

    def time_reconstruct_index_categorical(self, num_rows, dtype):
        self.mp._reconstruct_index_columns(
            df=self.df,
            key_indices=[("primary_key", str(dtype[1]))],
            columns=None,
            categories="primary_key",
            date_as_object=False,
        )



import inspect
import io
import logging
import os
import time
import warnings
from collections import namedtuple
from collections.abc import Callable, Iterable, Iterator, Sequence
from functools import wraps
from typing import (
    TYPE_CHECKING,
    Any,
    Union,
    cast,
)

import numpy as np
import pandas as pd
import pyarrow as pa
from minimalkv import KeyValueStore
from packaging import version
from pandas.api.types import is_datetime64_any_dtype

from plateau.core import naming
from plateau.core._compat import ARROW_GE_20, pandas_infer_string
from plateau.core.common_metadata import (
    SchemaWrapper,
    make_meta,
    normalize_column_order,
    read_schema_metadata,
    validate_compatible,
)
from plateau.core.docs import default_docs
from plateau.core.index import (
    ExplicitSecondaryIndex,
    IndexBase,
    merge_indices as merge_indices_algo,
)
from plateau.core.naming import get_partition_file_prefix
from plateau.core.partition import Partition
from plateau.core.typing import StoreInput
from plateau.core.urlencode import decode_key, quote_indices
from plateau.core.utils import ensure_store, ensure_string_type, verify_metadata_version
from plateau.core.uuid import gen_uuid
from plateau.io_components.utils import align_categories
from plateau.serialization import (
    DataFrameSerializer,
    PredicatesType,
    default_serializer,
    filter_df_from_predicates,
)

if TYPE_CHECKING:
    import numpy.typing as npt

LOGGER = logging.getLogger(__name__)

PANDAS_LT_2 = version.parse(pd.__version__) < version.parse("2")

SINGLE_TABLE = "table"

_Literal = namedtuple("_Literal", ["column", "op", "value"])
_SplitPredicate = namedtuple("_SplitPredicate", ["key_part", "content_part"])

_METADATA_SCHEMA = {
    "partition_label": np.dtype("O") if not pandas_infer_string() else str,
    "row_group_id": np.dtype(int),
    "row_group_compressed_size": np.dtype(int),
    "row_group_uncompressed_size": np.dtype(int),
    "number_rows_total": np.dtype(int),
    "number_row_groups": np.dtype(int),
    "serialized_size": np.dtype(int),
    "number_rows_per_row_group": np.dtype(int),
}

MetaPartitionInput = Union[pd.DataFrame, Sequence, "MetaPartition"] | None


def _predicates_to_named(predicates):
    if predicates is None:
        return None
    return [[_Literal(*x) for x in conjunction] for conjunction in predicates]


def _combine_predicates(predicates, logical_conjunction):
    if not logical_conjunction:
        return predicates
    if predicates is None:
        return [logical_conjunction]
    combined_predicates = []
    for conjunction in predicates:
        new_conjunction = conjunction[:]
        for literal in logical_conjunction:
            new_conjunction.append(literal)
        combined_predicates.append(new_conjunction)
    return combined_predicates


def _initialize_store_for_metapartition(method, method_args, method_kwargs):
    for store_variable in ["store", "storage"]:
        if store_variable in method_kwargs:
            method_kwargs[store_variable] = ensure_store(method_kwargs[store_variable])
        else:
            method = cast(object, method)
            args = inspect.getfullargspec(method).args

            if store_variable in args:
                ix = args.index(store_variable)
                # reduce index since the argspec and method_args start counting differently due to self
                ix -= 1
                instantiated_store = ensure_store(method_args[ix])
                new_args = []
                for ix_method, arg in enumerate(method_args):
                    if ix_method != ix:
                        new_args.append(arg)
                    else:
                        new_args.append(instantiated_store)
                method_args = tuple(new_args)

    return method_args, method_kwargs


def _apply_to_list(method):
    """Decorate a MetaPartition method to act upon the internal list of
    metapartitions.

    The methods must return a MetaPartition object!
    """

    @wraps(method)
    def _impl(self, *method_args, **method_kwargs):
        if not isinstance(self, MetaPartition):
            raise TypeError("Type unknown %s", type(self))

        result = self.as_sentinel()
        if len(self) == 0:
            raise RuntimeError("Invalid MetaPartition. No sub-partitions to act upon.")

        # Look whether there is a `store` in the arguments and instantiate it
        # this way we avoid multiple HTTP pools
        method_args, method_kwargs = _initialize_store_for_metapartition(
            method, method_args, method_kwargs
        )
        method_return = None  # declare for mypy
        if (len(self) == 1) and (self.label is None):
            result = method(self, *method_args, **method_kwargs)
        else:
            for mp in self:
                method_return = method(mp, *method_args, **method_kwargs)
                if not isinstance(method_return, MetaPartition):
                    raise ValueError(
                        f"Method {method.__name__} did not return a MetaPartition but {type(method_return)}"
                    )
                if method_return.is_sentinel:
                    result = method_return
                else:
                    for mp in method_return:
                        result = result.add_metapartition(mp, schema_validation=False)
        if not isinstance(result, MetaPartition):
            raise ValueError(
                f"Result for method {method.__name__} is not a `MetaPartition` but {type(method_return)}"
            )
        return result

    return _impl


class MetaPartitionIterator(Iterator):
    def __init__(self, metapartition):
        self.metapartition = metapartition
        self.position = 0

    def __iter__(self):
        return self

    def __next__(self):
        current = self.metapartition
        if len(current) == 1:
            if current.label is None:
                raise StopIteration()

        if self.position >= len(current.metapartitions):
            raise StopIteration()
        else:
            mp_dict = current.metapartitions[self.position]
            # These are global attributes, i.e. the nested metapartitions do not carry these and need
            # to be added here
            mp_dict["metadata_version"] = current.metadata_version
            mp_dict["schema"] = current.schema
            mp_dict["partition_keys"] = current.partition_keys
            mp_dict["logical_conjunction"] = current.logical_conjunction
            mp_dict["table_name"] = current.table_name
            self.position += 1
            return MetaPartition.from_dict(mp_dict)

    next = __next__  # Python 2


class MetaPartition(Iterable):
    """Wrapper for plateau partition which includes additional information
    about the parent dataset."""

    def __init__(
        self,
        label: str | None,
        file: str | None = None,
        table_name: str = SINGLE_TABLE,
        data: pd.DataFrame | None = None,
        indices: dict[Any, Any] | None = None,
        metadata_version: int | None = None,
        schema: SchemaWrapper | None = None,
        partition_keys: Sequence[str] | None = None,
        logical_conjunction: list[tuple[Any, str, Any]] | None = None,
    ):
        """Initialize the :mod:`plateau.io` base class MetaPartition.

        The `MetaPartition` is used as a wrapper around the plateau
        `Partition` and primarily deals with dataframe manipulations,
        in- and output to store.

        The :class:`plateau.io_components.metapartition` is immutable, i.e. all member
        functions will return a new MetaPartition object where the new
        attribute is changed

        Parameters
        ----------
        label
            partition label
        files
            A dictionary with references to the files in store where the
            keys represent file labels and the keys file prefixes.
        metadata
            The metadata of the partition
        data
            A dictionary including the materialized in-memory DataFrames
            corresponding to the file references in `files`.
        indices
            plateau index dictionary,
        metadata_version
        table_meta
            The dataset table schemas
        partition_keys
            The dataset partition keys
        logical_conjunction
            A logical conjunction to assign to the MetaPartition. By assigning
            this, the MetaPartition will only be able to load data respecting
            this conjunction.
        """

        if metadata_version is None:
            self.metadata_version = naming.DEFAULT_METADATA_VERSION
        else:
            self.metadata_version = metadata_version
        verify_metadata_version(self.metadata_version)
        self.schema = schema
        self.table_name = table_name
        if data is not None and schema is None:
            self.schema = make_meta(
                data, origin=f"{table_name}/{label}", partition_keys=partition_keys
            )

        indices = indices or {}
        for column, index_dct in indices.items():
            if isinstance(index_dct, dict):
                indices[column] = ExplicitSecondaryIndex(
                    column=column, index_dct=index_dct
                )
        self.logical_conjunction = logical_conjunction
        self.metapartitions = [
            {
                "label": label,
                "data": data,
                "file": file or None,
                "indices": indices,
                "logical_conjunction": logical_conjunction,
            }
        ]
        self.partition_keys = partition_keys or []

    def __repr__(self):
        if len(self.metapartitions) > 1:
            label = f"NESTED ({len(self.metapartitions)})"
        else:
            label = self.label
        return f"<{self.__class__.__name__} v{self.metadata_version} | {label} >"

    def __len__(self):
        return len(self.metapartitions)

    def __iter__(self):
        return MetaPartitionIterator(self)

    def __getitem__(self, label):
        for mp in self:
            if mp.label == label:
                return mp
        raise KeyError(f"Metapartition doesn't contain partition `{label}`")

    @property
    def data(self):
        if len(self.metapartitions) > 1:
            raise AttributeError(
                "Accessing `data` attribute is not allowed while nested"
            )
        assert isinstance(self.metapartitions[0], dict), self.metapartitions
        return self.metapartitions[0]["data"]

    @property
    def file(self) -> str:
        if len(self.metapartitions) > 1:
            raise AttributeError(
                "Accessing `files` attribute is not allowed while nested"
            )
        return cast(str, self.metapartitions[0]["file"])

    @property
    def is_sentinel(self) -> bool:
        return len(self.metapartitions) == 1 and self.label is None

    @property
    def label(self) -> str:
        if len(self.metapartitions) > 1:
            raise AttributeError(
                "Accessing `label` attribute is not allowed while nested"
            )
        assert isinstance(self.metapartitions[0], dict), self.metapartitions[0]
        return cast(str, self.metapartitions[0]["label"])

    @property
    def indices(self):
        if len(self.metapartitions) > 1:
            raise AttributeError(
                "Accessing `indices` attribute is not allowed while nested"
            )
        return self.metapartitions[0]["indices"]

    @property
    def partition(self) -> Partition:
        return Partition(label=self.label, files={self.table_name: self.file})

    def __eq__(self, other):
        if not isinstance(other, MetaPartition):
            return False

        if self.metadata_version != other.metadata_version:
            return False

        if self.schema is not None and not self.schema.equals(other.schema):
            return False

        if len(self.metapartitions) != len(other.metapartitions):
            return False

        # In the case both MetaPartitions are nested, we need to ensure a match
        # for all sub-partitions.
        # Since the label is unique, this can be used as a distinguishing key to sort and compare
        # the nested metapartitions.
        if len(self.metapartitions) > 1:
            for mp_self, mp_other in zip(
                sorted(self.metapartitions, key=lambda x: x["label"]),  # type: ignore
                sorted(other.metapartitions, key=lambda x: x["label"]),  # type: ignore
                strict=False,  # type: ignore
            ):
                if mp_self == mp_other:
                    continue
                # If a single metapartition does not match, the whole object is considered different
                return False
            return True

        # This is unnested only

        if self.label != other.label:
            return False

        if self.file != other.file:
            return False

        if self.data is not None and not self.data.equals(other.data):
            return False

        return True

    @staticmethod
    def from_partition(
        partition: Partition,
        data: pd.DataFrame | None = None,
        indices: dict | None = None,
        metadata_version: int | None = None,
        schema: SchemaWrapper | None = None,
        partition_keys: list[str] | None = None,
        logical_conjunction: list[tuple[Any, str, Any]] | None = None,
        table_name: str = SINGLE_TABLE,
    ):
        """Transform a plateau :class:`~plateau.core.partition.Partition` into
        a :class:`~plateau.io_components.metapartition.MetaPartition`.

        Parameters
        ----------
        partition
            The plateau partition to be wrapped
        data
            A dictionaries with materialised :class:`~pandas.DataFrame`
        indices : dict
            The index dictionary of the dataset
        schema
            Type metadata for each table, optional
        metadata_version
        partition_keys
            A list of the primary partition keys

        Returns
        -------
        :class:`~plateau.io_components.metapartition.MetaPartition`
        """
        return MetaPartition(
            label=partition.label,
            file=partition.files[table_name],
            data=data,
            indices=indices,
            metadata_version=metadata_version,
            schema=schema,
            partition_keys=partition_keys,
            logical_conjunction=logical_conjunction,
            table_name=table_name,
        )

    def add_metapartition(
        self,
        metapartition: "MetaPartition",
        schema_validation: bool = True,
    ):
        """Adds a metapartition to the internal list structure to enable batch
        processing.

        Parameters
        ----------
        metapartition
            The MetaPartition to be added.
        schema_validation
            If True (default), ensure that the `table_meta` of both `MetaPartition` objects are the same
        """
        if self.is_sentinel:
            return metapartition

        existing_label = [mp_["label"] for mp_ in self.metapartitions]

        if any(mp_["label"] in existing_label for mp_ in metapartition.metapartitions):
            raise RuntimeError(
                "Duplicate labels for nested metapartitions are not allowed!"
            )
        schema = metapartition.schema

        if schema_validation and schema:
            # This ensures that only schema-compatible metapartitions can be nested
            # The returned schema by validate_compatible is the reference schema with the most
            # information, i.e. the fewest null columns
            schema = validate_compatible([self.schema, metapartition.schema])

        new_object = MetaPartition(
            label="NestedMetaPartition",
            metadata_version=metapartition.metadata_version,
            schema=schema,
            partition_keys=metapartition.partition_keys or None,
            logical_conjunction=metapartition.logical_conjunction or None,
            table_name=metapartition.table_name,
        )

        # Add metapartition information to the new object
        new_metapartitions = self.metapartitions.copy()
        new_metapartitions.extend(metapartition.metapartitions.copy())
        new_object.metapartitions = new_metapartitions

        return new_object

    @staticmethod
    def from_dict(dct):
        """Create a :class:`~plateau.io_components.metapartition.MetaPartition`
        from a dictionary.

        Parameters
        ----------
        dct : dict
            Dictionary containing constructor arguments as keys

        Returns
        -------
        """
        return MetaPartition(
            label=dct["label"],
            file=dct.get("file", None),
            data=dct.get("data", None),
            table_name=dct.get("table_name", SINGLE_TABLE),
            indices=dct.get("indices", {}),
            metadata_version=dct.get("metadata_version", None),
            schema=dct.get("schema", None),
            partition_keys=dct.get("partition_keys", None),
            logical_conjunction=dct.get("logical_conjunction", None),
        )

    def to_dict(self):
        return {
            "label": self.label,
            "file": self.file,
            "data": self.data,
            "indices": self.indices,
            "metadata_version": self.metadata_version,
            "schema": self.schema,
            "partition_keys": self.partition_keys,
            "logical_conjunction": self.logical_conjunction,
            "table_name": self.table_name,
        }

    @_apply_to_list
    def remove_dataframes(self):
        """Remove all dataframes from the metapartition in memory."""
        return self.copy(data=None)

    def _split_predicates_in_index_and_content(self, predicates):
        """Split a list of predicates in the parts that can be resolved by the
        partition columns and the ones that are persisted in the data file."""
        # Predicates are split in this function into the parts that apply to
        # the partition key columns `key_part` and the parts that apply to the
        # contents of the file `content_part`.
        split_predicates = []
        has_index_condition = False
        for conjunction in predicates:
            key_part = []
            content_part = []
            for literal in conjunction:
                if literal.column in self.partition_keys:
                    has_index_condition = True
                    key_part.append(literal)
                else:
                    content_part.append(literal)
            split_predicates.append(_SplitPredicate(key_part, content_part))
        return split_predicates, has_index_condition

    def _apply_partition_key_predicates(self, indices, split_predicates):
        """Apply the predicates to the partition_key columns and return the
        remaining predicates that should be pushed to the DataFrame
        serialiser."""
        # Construct a single line DF with the partition columns
        schema = self.schema
        if schema is None:
            raise ValueError("schema needs to be loaded before applying predicates.")
        index_df_dct = {}
        for column, value in indices:
            pa_dtype = schema[schema.get_field_index(column)].type
            value = IndexBase.normalize_value(pa_dtype, value)
            if pa.types.is_date(pa_dtype):
                index_df_dct[column] = (
                    pd.Series(
                        pd.to_datetime([value], infer_datetime_format=True)
                    ).dt.date
                    # As of Pandas 2.0 infer_datetime_format is deprecated and is a no-op.
                    if PANDAS_LT_2
                    else pd.Series(pd.to_datetime([value])).dt.date
                )
            elif not ARROW_GE_20 and pa.types.is_large_string(pa_dtype):
                index_df_dct[column] = pd.Series([value])
            else:
                dtype = pa_dtype.to_pandas_dtype()
                index_df_dct[column] = pd.Series([value], dtype=dtype)
        index_df = pd.DataFrame(index_df_dct)

        filtered_predicates = []
        # We assume that indices on the partition level have been filtered out already in `dispatch_metapartitions`.
        # `filtered_predicates` should only contain predicates that can be evaluated on parquet level
        for conjunction in split_predicates:
            predicates = [conjunction.key_part]
            if (
                len(conjunction.key_part) == 0
                or len(
                    filter_df_from_predicates(
                        index_df, predicates, strict_date_types=True
                    )
                )
                > 0
            ):
                if len(conjunction.content_part) > 0:
                    filtered_predicates.append(conjunction.content_part)
                else:
                    # A condititon applies to the whole DataFrame, so we need to
                    # load all data.
                    return None
        return filtered_predicates

    @default_docs
    @_apply_to_list
    def load_dataframes(
        self,
        store: KeyValueStore,
        columns: Sequence[str] | None = None,
        predicate_pushdown_to_io: bool = True,
        categoricals: Sequence[str] | None = None,
        dates_as_object: bool = True,
        predicates: PredicatesType = None,
    ) -> "MetaPartition":
        """Load the dataframes of the partitions from store into memory.

        Parameters
        ----------
        tables
            If a list is supplied, only the given tables of the partition are
            loaded. If the given table does not exist it is ignored.

            Examples

            .. code::

                >>> part = MetaPartition(
                ...     label='part_label'
                ...     files={
                ...         'core': 'core_key_in_store',
                ...         'helper': 'helper_key_in_store'
                ...     }
                ...  )
                >>> part.data
                    {}
                >>> part = part.load_dataframes(store, ['core'])
                >>> part.data
                    {
                        'core': pd.DataFrame()
                    }
        """

        if categoricals is None:
            categoricals = []
        if not dates_as_object:
            warnings.warn(
                "The argument `date_as_object` is set to False. This argument will be deprecated and the future behaviour will be as if the paramere was set to `True`. Please migrate your code accordingly ahead of time.",
                DeprecationWarning,
                stacklevel=2,
            )

        LOGGER.debug("Loading internal dataframes of %s", self.label)
        if not self.file:
            # This used to raise, but the specs do not require this, so simply do a no op
            LOGGER.debug("Partition %s is empty and has no data.", self.label)
            return self
        predicates = _combine_predicates(predicates, self.logical_conjunction)
        predicates = _predicates_to_named(predicates)

        dataset_uuid, _, indices, _ = decode_key(self.file)

        # In case the columns only refer to the partition indices, we need to load at least a single column to
        # determine the length of the required dataframe.
        table_columns_to_io = columns

        filtered_predicates = predicates

        self = self.load_schema(dataset_uuid=dataset_uuid, store=store)

        # Filter predicates that would apply to this partition and remove the partition columns
        if predicates:
            # Check if there are predicates that match to the partition columns.
            # For these we need to check if the partition columns already falsify
            # the conditition.
            #
            # We separate these predicates into their index and their Parquet part.
            (
                split_predicates,
                has_index_condition,
            ) = self._split_predicates_in_index_and_content(predicates)

            filtered_predicates = []
            if has_index_condition:
                filtered_predicates = self._apply_partition_key_predicates(
                    indices, split_predicates
                )
            else:
                filtered_predicates = [pred.content_part for pred in split_predicates]

        # Remove partition_keys from table_columns_to_io
        if self.partition_keys and table_columns_to_io is not None:
            keys_to_remove = set(self.partition_keys) & set(table_columns_to_io)
            # This is done to not change the ordering of the list
            table_columns_to_io = [
                c for c in table_columns_to_io if c not in keys_to_remove
            ]

        start = time.time()
        df = DataFrameSerializer.restore_dataframe(
            key=self.file,
            store=store,
            columns=table_columns_to_io,
            categories=categoricals,
            predicate_pushdown_to_io=predicate_pushdown_to_io,
            predicates=filtered_predicates,
            date_as_object=dates_as_object,
        )
        LOGGER.debug(
            "Loaded dataframe %s in %s seconds.", self.file, time.time() - start
        )
        # Metadata version >=4 parse the index columns and add them back to the dataframe

        df = self._reconstruct_index_columns(
            df=df,
            key_indices=indices,
            columns=columns,
            categories=categoricals,
            date_as_object=dates_as_object,
        )

        df.columns = df.columns.map(ensure_string_type)
        if columns is not None:
            # TODO: When the write-path ensures that all partitions have the same column set, this check can be
            #       moved before `DataFrameSerializer.restore_dataframe`. At the position of the current check we
            #       may want to double check the columns of the loaded DF and raise an exception indicating an
            #       inconsistent dataset state instead.
            missing_cols = set(columns).difference(df.columns)
            if missing_cols:
                raise ValueError(
                    "Columns cannot be found in stored dataframe: {}".format(
                        ", ".join(sorted(missing_cols))
                    )
                )

            if list(df.columns) != columns:
                df = df.reindex(columns=columns, copy=False)

        return self.copy(data=df)

    @_apply_to_list
    def load_schema(self, store: StoreInput, dataset_uuid: str) -> "MetaPartition":
        """Loads all table metadata in memory and stores it under the `tables`
        attribute."""

        if self.schema is None:
            store = ensure_store(store)
            self.schema = read_schema_metadata(
                dataset_uuid=dataset_uuid, store=store, table=self.table_name
            )
        return self

    def _reconstruct_index_columns(
        self, df, key_indices, columns, categories, date_as_object
    ):
        if len(key_indices) == 0:
            return df

        original_columns = list(df.columns)
        zeros: npt.NDArray[np.int_] = np.zeros(len(df), dtype=int)
        schema = self.schema
        if schema is None:
            raise ValueError("Cannot reconstruct indices before the schema is loaded.")

        # One of the few places `inplace=True` makes a significant difference
        df.reset_index(drop=True, inplace=True)

        index_names = [primary_key for primary_key, _ in key_indices]
        # The index might already be part of the dataframe which is recovered from the parquet file.
        # In this case, still use the reconstructed index col to have consistent index columns behavior.
        # In this case the column in part of `original_columns` and must be removed to avoid duplication
        # in the column axis
        cleaned_original_columns = [
            orig for orig in original_columns if orig not in index_names
        ]
        if cleaned_original_columns != original_columns:
            # indexer call is slow, so only do that if really necessary
            df = df.reindex(columns=cleaned_original_columns, copy=False)

        pos = 0
        for primary_key, value in key_indices:
            # If there are predicates, don't reconstruct the index if it wasn't requested
            if columns is not None and primary_key not in columns:
                continue

            pa_dtype = schema.field(primary_key).type
            if not ARROW_GE_20 and pa.types.is_large_string(pa_dtype):
                if pandas_infer_string():
                    dtype = "string[pyarrow_numpy]"
                else:
                    dtype = "object"
                dtype = pd.api.types.pandas_dtype(dtype)
            else:
                dtype = pa_dtype.to_pandas_dtype()
            convert_to_date = False
            if date_as_object and pa_dtype in [pa.date32(), pa.date64()]:
                convert_to_date = True

            if isinstance(dtype, type):
                value = dtype(value)
            elif isinstance(dtype, np.dtype):
                if is_datetime64_any_dtype(dtype):
                    # Coerce all datetime64 units to nanoseconds to maintain consistency with serializers.
                    value = (
                        pd.Timestamp(value)
                        if PANDAS_LT_2
                        else pd.Timestamp(value).as_unit("ns")
                    )
                else:
                    value = dtype.type(value)
            else:
                raise RuntimeError(
                    f"Unexpected object encountered: ({dtype}, {type(dtype)})"
                )
            if categories and primary_key in categories:
                if convert_to_date:
                    cats = pd.Series(value).dt.date
                else:
                    cats = [value]
                value = pd.Categorical.from_codes(zeros, categories=cats)
            else:
                if convert_to_date:
                    value = pd.Timestamp(value).to_pydatetime().date()
            df.insert(pos, primary_key, value)
            pos += 1

        return df

    @_apply_to_list
    def validate_schema_compatible(
        self, store: StoreInput, dataset_uuid: str
    ) -> "MetaPartition":
        """Validates that the currently held DataFrames match the schema of the
        existing dataset.

        Parameters
        ----------
        store
            If it is a function, the result of calling it must be a KeyValueStore.
        dataset_uuid
            The dataset UUID the partition will be assigned to
        """

        # Load the reference meta of the existing dataset. Using the built-in
        # `load_all_table_meta` would not be helpful here as it would be a no-op
        # as we have already loaded the meta from the input DataFrame.
        store = ensure_store(store)
        reference_meta = read_schema_metadata(
            dataset_uuid=dataset_uuid, store=store, table=self.table_name
        )
        try:
            validate_compatible([self.schema, reference_meta])
        except ValueError as e:
            raise ValueError(
                f"Schemas for dataset '{dataset_uuid}' are not compatible!\n\n{e}"
            ) from e
        return self

    @_apply_to_list
    def store_dataframes(
        self,
        store: StoreInput,
        dataset_uuid: str,
        df_serializer: DataFrameSerializer | None = None,
    ) -> "MetaPartition":
        """Stores all dataframes of the MetaPartitions and registers the saved
        files under the `files` attribute. The dataframe itself is deleted from
        memory.

        Parameters
        ----------
        store
            If it is a function, the result of calling it must be a KeyValueStore.
        dataset_uuid
            The dataset UUID the partition will be assigned to
        df_serializer
            Serialiser to be used to store the dataframe
        Returns
        -------
        MetaPartition
        """
        df_serializer = (
            df_serializer if df_serializer is not None else default_serializer()
        )
        actual_store = ensure_store(store)

        key = get_partition_file_prefix(
            partition_label=self.label,
            dataset_uuid=dataset_uuid,
            metadata_version=self.metadata_version,
            table=self.table_name,
        )
        if self.data is not None:
            df = self.data
            try:
                file = df_serializer.store(actual_store, key, df)
            except Exception as exc:
                try:
                    if isinstance(df, pd.DataFrame):
                        buf = io.StringIO()
                        df.info(buf=buf)
                        LOGGER.error(
                            "Writing dataframe failed.\n%s\n%s\n%s",
                            exc,
                            buf.getvalue(),
                            df.head(),
                        )
                    else:
                        LOGGER.error("Storage of dask dataframe failed.")
                        pass
                finally:
                    raise

            new_metapartition = self.copy(file=file, data=None)

            return new_metapartition
        else:
            return self

    @_apply_to_list
    def apply(
        self,
        func: Callable,
        type_safe: bool = False,
    ) -> "MetaPartition":
        """Applies a given function to all dataframes of the MetaPartition.

        Parameters
        ----------
        func
            A callable accepting and returning a :class:`pandas.DataFrame`
        uuid :
            The changed dataset is assigned a new UUID.
        type_safe
            If the transformation is type-safe, optimizations can be applied
        """

        new_data = func(self.data)

        if type_safe:
            new_schema = self.schema
        else:
            new_schema = make_meta(
                new_data,
                origin=self.label,
                partition_keys=self.partition_keys,
            )
        return self.copy(data=new_data, schema=new_schema)

    def as_sentinel(self):
        """"""
        return MetaPartition(
            None,
            metadata_version=self.metadata_version,
            partition_keys=self.partition_keys,
        )

    def copy(self, **kwargs):
        """Creates a shallow copy where the kwargs overwrite existing
        attributes."""

        def _renormalize_meta(meta):
            if "partition_keys" in kwargs and meta is not None:
                pk = kwargs["partition_keys"]
                return normalize_column_order(meta, pk)
            else:
                return meta

        metapartitions = kwargs.get("metapartitions", None) or []
        metapartitions.extend(self.metapartitions)
        if len(metapartitions) > 1:
            first_mp = metapartitions.pop()
            mp_parent = MetaPartition(
                label=first_mp.get("label"),
                file=first_mp.get("file"),
                data=first_mp.get("data"),
                indices=first_mp.get("indices"),
                metadata_version=self.metadata_version,
                schema=_renormalize_meta(kwargs.get("schema", self.schema)),
                partition_keys=kwargs.get("partition_keys", self.partition_keys),
                logical_conjunction=kwargs.get(
                    "logical_conjunction", self.logical_conjunction
                ),
                table_name=kwargs.get("table_name", self.table_name),
            )
            for mp in metapartitions:
                mp_parent = mp_parent.add_metapartition(
                    MetaPartition(
                        label=mp.get("label"),
                        file=mp.get("file"),
                        data=mp.get("data"),
                        indices=mp.get("indices"),
                        metadata_version=self.metadata_version,
                        schema=_renormalize_meta(kwargs.get("schema", self.schema)),
                        partition_keys=kwargs.get(
                            "partition_keys", self.partition_keys
                        ),
                        logical_conjunction=kwargs.get(
                            "logical_conjunction", self.logical_conjunction
                        ),
                        table_name=kwargs.get("table_name", self.table_name),
                    ),
                    schema_validation=False,
                )
            return mp_parent
        else:
            mp = MetaPartition(
                label=kwargs.get("label", self.label),
                file=kwargs.get("file", self.file),
                data=kwargs.get("data", self.data),
                indices=kwargs.get("indices", self.indices),
                metadata_version=kwargs.get("metadata_version", self.metadata_version),
                schema=_renormalize_meta(kwargs.get("schema", self.schema)),
                partition_keys=kwargs.get("partition_keys", self.partition_keys),
                logical_conjunction=kwargs.get(
                    "logical_conjunction", self.logical_conjunction
                ),
                table_name=kwargs.get("table_name", self.table_name),
            )
            return mp

    @_apply_to_list
    def build_indices(self, columns: Iterable[str]):
        """This builds the indices for this metapartition for the given
        columns. The indices for the passed columns are rebuilt, so existing
        index entries in the metapartition are overwritten.

        :param columns: A list of columns from which the indices over
            all dataframes in the metapartition are overwritten
        :return: self
        """
        if self.label is None:
            return self

        new_indices = {}
        for col in columns:
            possible_values: set[str] = set()

            df = self.data
            if not self.is_sentinel and col not in df:
                raise RuntimeError(
                    f"Column `{col}` could not be found in the partition `{self.label}` Please check for any typos and validate your dataset."
                )

            possible_values = possible_values | set(df[col].dropna().unique())

            if self.schema is not None:
                dtype = self.schema.field(col).type
            else:
                dtype = None

            new_index = ExplicitSecondaryIndex(
                column=col,
                index_dct={value: [self.label] for value in possible_values},
                dtype=dtype,
            )
            if (col in self.indices) and self.indices[col].loaded:
                new_indices[col] = self.indices[col].update(new_index)
            else:
                new_indices[col] = new_index

        return self.copy(indices=new_indices)

    @_apply_to_list
    def partition_on(self, partition_on: str | Sequence[str]):
        """Partition all dataframes assigned to this MetaPartition according
        the the given columns.

        If the MetaPartition object contains index information, the information is split in such a way that they
        reference the new partitions.

        In case a requested partition column is not existent in **all** tables, a KeyError is raised.

        All output partitions are re-assigned labels encoding the partitioned columns (urlencoded)

        Examples::

            >>> import pandas as pd
            >>> from plateau.io_components.metapartition import MetaPartition
            >>> mp = MetaPartition(
            ...     label='partition_label',
            ...     data={
            ...         "Table1": pd.DataFrame({
            ...             'P': [1, 2, 1, 2],
            ...             'L': [1, 1, 2, 2]
            ...         })
            ...     }
            ... )
            >>> repartitioned_mp = mp.partition_on(['P', 'L'])
            >>> assert [mp["label"] for mp in repartitioned_mp.metapartitions] == [
            ...     "P=1/L=1/partition_label",
            ...     "P=1/L=2/partition_label",
            ...     "P=2/L=1/partition_label",
            ...     "P=2/L=2/partition_label"
            ... ]

        Parameters
        ----------
        partition_on
        """
        if partition_on == self.partition_keys:
            return self

        for partition_column in partition_on:
            if partition_column in self.indices:
                raise ValueError(
                    "Trying to `partition_on` on a column with an explicit index!"
                )
        if self.is_sentinel:
            return self.copy(partition_keys=partition_on)
        else:
            new_mp = self.as_sentinel().copy(
                partition_keys=partition_on,
                schema=normalize_column_order(self.schema, partition_on),
            )

        if isinstance(partition_on, str):
            partition_on = [partition_on]
        partition_on = self._ensure_compatible_partitioning(partition_on)

        new_data = self._partition_data(partition_on)

        for label, data in new_data.items():
            tmp_mp = MetaPartition(
                label=label,
                file=self.file,
                data=data,
                metadata_version=self.metadata_version,
                indices={},
                schema=normalize_column_order(self.schema, partition_on).with_origin(
                    f"{label}"
                ),
                partition_keys=partition_on,
                table_name=self.table_name,
            )
            new_mp = new_mp.add_metapartition(tmp_mp, schema_validation=False)
        if self.indices:
            new_mp = new_mp.build_indices(columns=self.indices.keys())
        return new_mp

    def _ensure_compatible_partitioning(self, partition_on):
        if (
            not self.partition_keys
            or self.partition_keys
            and (len(partition_on) >= len(self.partition_keys))
            and (self.partition_keys == partition_on[: len(self.partition_keys)])
        ):
            return partition_on[len(self.partition_keys) :]
        else:
            raise ValueError(
                "Incompatible partitioning encountered. `partition_on` needs to include the already "
                "existing partition keys and must preserve their order.\n"
                f"Current partition keys: `{self.partition_keys}`\n"
                f"Partition on called with: `{partition_on}`"
            )

    def _partition_data(self, partition_on):
        existing_indices, base_label = cast(
            tuple[list, str], decode_key(f"uuid/table/{self.label}")[2:]
        )
        dct: dict[str, Any] = {}
        df = self.data

        # Check that data sizes do not change. This might happen if the
        # groupby below drops data, e.g. nulls
        size_after = 0
        size_before = len(df)

        # Implementation from pyarrow
        # See https://github.com/apache/arrow/blob/b33dfd9c6bd800308bb1619b237dbf24dea159be/python/pyarrow/parquet.py#L1030  # noqa: E501

        # column sanity checks
        data_cols = set(df.columns).difference(partition_on)
        missing_po_cols = set(partition_on).difference(df.columns)
        if missing_po_cols:
            raise ValueError(
                "Partition column(s) missing: {}".format(
                    ", ".join(sorted(missing_po_cols))
                )
            )
        if len(data_cols) == 0:
            raise ValueError("No data left to save outside partition columns")

        # To be aligned with open source tooling we drop the index columns and recreate
        # them upon reading as it is done by fastparquet and pyarrow
        partition_keys = [df[col] for col in partition_on]

        # # The handling of empty dfs is not part of the arrow implementation
        # if df.empty:
        #     return {}

        data_df = df.drop(partition_on, axis="columns")
        for value, group in data_df.groupby(
            by=partition_keys, observed=True, sort=False
        ):
            partitioning_info = []
            if pd.api.types.is_scalar(value):
                value = [value]
            if existing_indices:
                partitioning_info.extend(quote_indices(existing_indices))
            partitioning_info.extend(
                quote_indices(zip(partition_on, value, strict=False))
            )
            partitioning_info.append(base_label)
            new_label = "/".join(partitioning_info)

            if new_label not in dct:
                dct[new_label] = {}
            dct[new_label] = group
            size_after += len(group)

        if size_before != size_after:
            raise ValueError(
                f"Original dataframe size ({size_before} rows) does not "
                f"match new dataframe size ({size_after} rows). "
                f"Hint: you may see this if you are trying to use `partition_on` on a column with null values."
            )

        return dct

    @staticmethod
    def merge_indices(metapartitions):
        list_of_indices = []
        for mp in metapartitions:
            for sub_mp in mp:
                if sub_mp.indices:
                    list_of_indices.append(sub_mp.indices)
        return merge_indices_algo(list_of_indices)

    @staticmethod
    def _merge_labels(metapartitions, label_merger=None):
        # Use the shortest of available labels since this has to be the partition
        #  label prefix
        new_label = None
        # FIXME: This is probably not compatible with >= v3
        if label_merger is None:
            for mp in metapartitions:
                label = mp.label
                if new_label is None or len(label) < len(new_label):
                    new_label = label
                    continue
        else:
            new_label = label_merger([mp.label for mp in metapartitions])

        return new_label

    @staticmethod
    def concat_metapartitions(metapartitions, label_merger=None):
        LOGGER.debug("Concatenating metapartitions")

        new_metadata_version = -1
        data = []
        schema = []
        for mp in metapartitions:
            new_metadata_version = max(new_metadata_version, mp.metadata_version)
            data.append(mp.data)
            schema.append(mp.schema)
            # Don't care about the partition_keys. If we try to merge
            # MetaPartitions without alignment the schemas won't match.
            partition_keys = mp.partition_keys

        categoricals = [
            col
            for col, ser in data[0].items()
            if isinstance(ser.dtype, pd.CategoricalDtype)
        ]
        if categoricals:
            data = align_categories(data, categoricals)
        new_df = pd.concat(data)

        new_schema = validate_compatible(schema)

        new_label = MetaPartition._merge_labels(metapartitions, label_merger)

        new_mp = MetaPartition(
            label=new_label,
            data=new_df,
            metadata_version=new_metadata_version,
            schema=new_schema,
            partition_keys=partition_keys,
        )

        return new_mp

    @_apply_to_list
    def delete_from_store(
        self, dataset_uuid: Any, store: StoreInput
    ) -> "MetaPartition":
        store = ensure_store(store)
        # Delete data first
        store.delete(self.file)
        return self.copy(file=None, data=None)

    def get_parquet_metadata(self, store: StoreInput) -> pd.DataFrame:
        """Retrieve the parquet metadata for the MetaPartition. Especially
        relevant for calculating dataset statistics.

        Parameters
        ----------
        store
          A factory function providing a KeyValueStore
        table_name
          Name of the plateau table for which the statistics should be retrieved

        Returns
        -------
        pd.DataFrame
          A DataFrame with relevant parquet metadata
        """
        store = ensure_store(store)

        data = {}
        with store.open(self.file) as fd:  # type: ignore
            pq_metadata = pa.parquet.ParquetFile(fd).metadata

        data = {
            "partition_label": self.label,
            "serialized_size": pq_metadata.serialized_size,
            "number_rows_total": pq_metadata.num_rows,
            "number_row_groups": pq_metadata.num_row_groups,
            "row_group_id": [],
            "number_rows_per_row_group": [],
            "row_group_compressed_size": [],
            "row_group_uncompressed_size": [],
        }
        for rg_ix in range(pq_metadata.num_row_groups):
            rg = pq_metadata.row_group(rg_ix)
            data["row_group_id"].append(rg_ix)
            data["number_rows_per_row_group"].append(rg.num_rows)
            data["row_group_compressed_size"].append(rg.total_byte_size)
            data["row_group_uncompressed_size"].append(
                sum(
                    rg.column(col_ix).total_uncompressed_size
                    for col_ix in range(rg.num_columns)
                )
            )

        df = pd.DataFrame(data=data, columns=_METADATA_SCHEMA.keys())
        df = df.astype(_METADATA_SCHEMA)
        return df


def _unique_label(label_list):
    label = os.path.commonprefix(label_list)
    if len(label) == 0:
        label = "_".join(label_list)
    while len(label) > 0 and not label[-1].isalnum():
        label = label[:-1]
    return label


def partition_labels_from_mps(mps: list[MetaPartition]) -> list[str]:
    """Get a list of partition labels, flattening any nested meta partitions in
    the input and ignoring sentinels."""
    partition_labels = []
    for mp in mps:
        if len(mp) > 1:
            for nested_mp in mp:
                if not nested_mp.is_sentinel:
                    partition_labels.append(nested_mp.label)
        else:
            if not mp.is_sentinel:
                partition_labels.append(mp.label)
    return partition_labels


def parse_input_to_metapartition(
    obj: MetaPartitionInput,
    table_name: str = SINGLE_TABLE,
    metadata_version: int | None = None,
) -> MetaPartition:
    """Parses given user input and return a MetaPartition.

    The expected input is a :class:`pandas.DataFrame` or a list of
    :class:`pandas.DataFrame`.

    Every element of the list will be treated as a dedicated user input and will
    result in a physical file, if not specified otherwise.

    Parameters
    ----------
    obj
    table_name
        The table name assigned to the partitions
    metadata_version
        The plateau dataset specification version
    """

    if obj is None:
        obj = []
    if isinstance(obj, list):
        if len(obj) == 0:
            return MetaPartition(label=None, metadata_version=metadata_version)
        first_element = obj[0]
        mp = parse_input_to_metapartition(
            obj=first_element,
            metadata_version=metadata_version,
            table_name=table_name,
        )
        for mp_in in obj[1:]:
            mp = mp.add_metapartition(
                parse_input_to_metapartition(
                    obj=mp_in,
                    metadata_version=metadata_version,
                    table_name=table_name,
                )
            )
    elif isinstance(obj, pd.DataFrame):
        mp = MetaPartition(
            label=gen_uuid(),
            data=obj,
            metadata_version=metadata_version,
            table_name=table_name,
        )
    elif isinstance(obj, MetaPartition):
        return obj
    else:
        raise ValueError(
            f"Unexpected type during parsing encountered: ({type(obj)}, {obj})"
        )

    return mp



import contextlib

import distributed
import distributed.utils_test
import pytest
from distributed import Client

import plateau.core._time
from plateau.core.testing import cm_frozen_time

_client = None


@pytest.fixture(autouse=True, scope="session")
def setup_dask_distributed():
    """This fixture makes all dask tests effectively use distributed under the
    hood."""
    global _client
    with distributed.utils_test.cluster() as (scheduler, workers):
        _client = Client(scheduler["address"])
        yield


@contextlib.contextmanager
def cm_distributed_frozen_time():
    global _client
    assert _client is not None
    _client.run(_freeze_time_on_worker, plateau.core._time.datetime_now())
    try:
        yield
    finally:
        _client.run(_unfreeze_time_on_worker)


@pytest.fixture
def distributed_frozen_time(frozen_time):
    with cm_distributed_frozen_time():
        yield


@pytest.fixture
def frozen_time_em(frozen_time):
    with cm_distributed_frozen_time():
        yield


def _freeze_time_on_worker(freeze_time):
    # this runs on the distributed worker, initiated by cm_distributed_frozen_time
    # it runs the "enter" part of the cm_frozen_time context manager.
    # It saves the context manager to be able to call __exit__ on it later
    # in _unfreeze_time_on_worker.
    plateau.core._time._datetime_utcnow_orig = plateau.core._time.datetime_utcnow
    cm = cm_frozen_time(freeze_time)
    plateau.core._time._time_patcher = cm
    cm.__enter__()


def _unfreeze_time_on_worker():
    plateau.core._time._time_patcher.__exit__(None, None, None)
